import json
import sys
from pathlib import Path
from typing import Any
import shutil

import pytest
from pycivil.EXAStructural.checkable import CheckableCriteriaEnum
from pycivil.EXAStructural.rcrecsolver.srvRcSecCheck import (
    ModelInputRcSecCheck,
    ModelOutputRcSecCheck,
    RcSecRectSolver,
    SolverOptions,
    ReportOption
)
from pycivil.EXAStructuralCheckable.RcsRectangular import ThermalMapSolverIn

from pycivil.EXAStructural.codes import Code
from pycivil.EXAStructural.materials import Concrete, ConcreteModel, ConcreteSteel, SteelModel
from pycivil.EXAStructural.loads import ForcesOnSection
from pycivil.EXAStructural.sections import (
    RectangularShape,
    ShapePositionEnum,
    ConcreteSectionModel,
    SteelDisposerOnLine,
    SteelDisposerStirrup,
)


def read_json(file_name: str) -> dict[str, Any]:
    """Reads a JSON file placed in this directory."""
    base_path = Path(__file__).parent
    return json.loads(base_path.joinpath(file_name).read_text())


@pytest.mark.parametrize("file_name", [f"test_0{n}_RcSecSolver" for n in range(1, 5)])
def test_rc_sec_solver_by_hand(tmp_path: Path, file_name: str) -> None:
    solver = RcSecRectSolver()
    solver.setJobPath(str(tmp_path))

    jsonObjectStatic = read_json(f"{file_name}.json")
    shutil.copyfile(src=Path(__file__).parent.joinpath(f"{file_name}.json"), dst=tmp_path.joinpath(f"{file_name}.json"))

    iDataStatic = ModelInputRcSecCheck(**jsonObjectStatic)


    # TODO: skip this if server is unavailable
    if CheckableCriteriaEnum.SLU_NM_FIRE in iDataStatic.criteria:
        jsonObject = read_json(f"{file_name}_Thermal.json")
        iDataThermal = ThermalMapSolverIn(**jsonObject)
        solver.setModelInput(iDataThermal)
        solver.run(SolverOptions.THERMAL)

    solver.setModelInput(iDataStatic)
    solver.run(SolverOptions.STATIC)

    oData = solver.getModelOutput()
    assert isinstance(oData, ModelOutputRcSecCheck)
    for _i, c in enumerate(oData.results.resultsForCriteria):
        c.media = None

    # ------------------------------------------------
    # using this for generate not regression json file
    # ------------------------------------------------
    # outfile = Path(__file__).parent.joinpath(f"{file_name}_nr.json")
    # outfile.write_text(json.dumps(oData.model_dump(), indent=4))

    # Testing data persistence on file for oData
    data = ModelOutputRcSecCheck(**read_json(f"{file_name}_nr.json"))
    for _i, c in enumerate(data.results.resultsForCriteria):
        c.media = None

    assert data == oData

    solver.modelInputFile = f"{file_name}.json"
    shutil.copyfile(src=Path(__file__).parent.joinpath(f"{file_name}_nr.json"), dst=tmp_path.joinpath(f"{file_name}_nr.json"))
    solver.modelOutputFile = f"{file_name}_nr.json"
    solver.reportName = file_name
    assert solver.buildReport()

@pytest.mark.parametrize("file_name", [f"test_0{n}_RcSecSolver" for n in range(6, 12)])
def test_rc_sec_solver_front_end_case(tmp_path: Path, file_name: str) -> None:
    solver = RcSecRectSolver()
    solver.setJobPath(str(tmp_path))

    jsonObjectStatic = read_json(f"{file_name}.json")
    shutil.copyfile(src=Path(__file__).parent.joinpath(f"{file_name}.json"), dst=tmp_path.joinpath(f"{file_name}.json"))

    iDataStatic = ModelInputRcSecCheck(**jsonObjectStatic)

    # TODO: skip this if server is unavailable
    if CheckableCriteriaEnum.SLU_NM_FIRE in iDataStatic.criteria:
        jsonObject = read_json(f"{file_name}_Thermal.json")
        iDataThermal = ThermalMapSolverIn(**jsonObject)
        solver.setModelInput(iDataThermal)
        solver.run(SolverOptions.THERMAL)

    solver.setModelInput(iDataStatic)
    solver.run(SolverOptions.STATIC)

    oData = solver.getModelOutput()
    assert isinstance(oData, ModelOutputRcSecCheck)
    for _i, c in enumerate(oData.results.resultsForCriteria):
        c.media = None

    # ------------------------------------------------
    # using this for generate not regression json file
    # ------------------------------------------------
    # outfile = Path(__file__).parent.joinpath(f"{file_name}_nr.json")
    # outfile.write_text(json.dumps(oData.model_dump(), indent=4))

    # Testing data persistence on file for oData
    data = ModelOutputRcSecCheck(**read_json(f"{file_name}_nr.json"))
    for _i, c in enumerate(data.results.resultsForCriteria):
        c.media = None

    assert data == oData

    solver.modelInputFile = f"{file_name}.json"
    shutil.copyfile(src=Path(__file__).parent.joinpath(f"{file_name}_nr.json"), dst=tmp_path.joinpath(f"{file_name}_nr.json"))

    solver.modelOutputFile = f"{file_name}_nr.json"
    solver.reportName = file_name
    assert solver.buildReport()

# This test case build a input as in test_01_RcSecSolver
def test_rc_solver_cl_usability_001(tmp_path):

    f5 = ForcesOnSection(Fx=-200000.0, My= 200000000.0, Fz=200000.0, descr="force_5")
    f6 = ForcesOnSection(Fx= 200000.0, My=-200000000.0, Fz=100000.0, descr="force_6")
    f7 = ForcesOnSection(Fx=-500000.0, My= 350000000.0, Fz=200000.0, descr="force_7")
    f8 = ForcesOnSection(Fx= 100000.0, My=-200000000.0, Fz= 70000.0, descr="force_8")

    loads = [f5, f6, f7, f8]
    for l in loads:
        l.setLimitState("ultimate")

    code = Code(codeStr="NTC2018")
    concrete = Concrete(descr="Concrete NTC2018 - C25/30")
    concrete.setByCode(code, "C25/30")
    concrete.setEnvironment("not aggressive")
    concreteModel = ConcreteModel()
    concreteModel.fromMaterial(concrete)

    steel = ConcreteSteel(descr="Steel NTC2018 - B450C")
    steel.setByCode(code, "B450C")
    steel.setSensitivity("not sensitive")
    steelModel = SteelModel()
    steelModel.fromMaterial(steel)

    shape = RectangularShape(width=300, height=600, descr="Template RC Section")
    section = ConcreteSectionModel(concreteMat=concreteModel, steelMat=steelModel, shape=shape)

    rebarsOnTop = SteelDisposerOnLine(
        fromPos=ShapePositionEnum.MT,
        diameter=16,
        steelInterDistance=40,
        distanceFromPos=40,
        number=4
    )

    rebarsOnBot = SteelDisposerOnLine(
        fromPos=ShapePositionEnum.MB,
        diameter=24,
        steelInterDistance=40,
        distanceFromPos=40,
        number=4
    )
    section.disposerOnLine = [rebarsOnTop, rebarsOnBot]

    section_solver_input = ModelInputRcSecCheck(loads=loads, section=section)

    section_solver_input.criteria = [CheckableCriteriaEnum.SLU_NM]
    section_solver_input.loadsInCriteria = [[0, 1, 2, 3]]

    solver = RcSecRectSolver()
    solver.setJobPath(str(tmp_path))
    solver.setModelInput(section_solver_input)
    solver.run(SolverOptions.STATIC)
    solver.reportName = "test_rc_solver_cl_usability_001"
    solver.buildReport(opt=ReportOption.REPORT_FROM_RUN)

# This test case build a input as in test_02_RcSecSolver
def test_rc_solver_cl_usability_002(tmp_path):
    f1 = ForcesOnSection(Fx= 150000.0, My= 150000000.0, Fz=150000.0, descr="force_1", id=1)
    f2 = ForcesOnSection(Fx= 100000.0, My= 145000000.0, Fz=120000.0, descr="force_2", id=2)
    f3 = ForcesOnSection(Fx=-100000.0, My= 145000000.0, Fz=120000.0, descr="force_3", id=3)
    f4 = ForcesOnSection(Fx=-125000.0, My= 125000000.0, Fz=125000.0, descr="force_4", id=4)
    f5 = ForcesOnSection(Fx=-200000.0, My= 200000000.0, Fz=200000.0, descr="force_5", id=5)
    f6 = ForcesOnSection(Fx= 200000.0, My=-200000000.0, Fz=100000.0, descr="force_6", id=6)
    f7 = ForcesOnSection(Fx=-500000.0, My= 350000000.0, Fz=200000.0, descr="force_7", id=7)
    f8 = ForcesOnSection(Fx= 100000.0, My=-200000000.0, Fz= 70000.0, descr="force_8", id=8)

    f1.setLimitState("serviceability")
    f1.setFrequency("characteristic")
    f2.setLimitState("serviceability")
    f2.setFrequency("quasi-permanent")
    f3.setLimitState("serviceability")
    f3.setFrequency("quasi-permanent")
    f4.setLimitState("serviceability")
    f4.setFrequency("frequent")

    loads = [f1, f2, f3, f4, f5, f6, f7, f8]
    for l in [f5, f6, f7, f8]:
        l.setLimitState("ultimate")

    code = Code(codeStr="NTC2018")
    concrete = Concrete(descr="My concrete")
    concrete.setByCode(code, "C25/30")
    concrete.setEnvironment("not aggressive")
    concreteModel = ConcreteModel()
    concreteModel.fromMaterial(concrete)

    steel = ConcreteSteel(descr="My steel")
    steel.setByCode(code, "B450C")
    steel.setSensitivity("not sensitive")
    steelModel = SteelModel()
    steelModel.fromMaterial(steel)

    shape = RectangularShape(width=300, height=600, descr="Template RC Section")
    section = ConcreteSectionModel(concreteMat=concreteModel, steelMat=steelModel, shape=shape)

    rebarsOnTop = SteelDisposerOnLine(
        fromPos=ShapePositionEnum.MT,
        diameter=20,
        steelInterDistance=40,
        distanceFromPos=40,
        number=4
    )

    rebarsOnBot = SteelDisposerOnLine(
        fromPos=ShapePositionEnum.MB,
        diameter=20,
        steelInterDistance=40,
        distanceFromPos=40,
        number=4
    )
    section.disposerOnLine = [rebarsOnTop, rebarsOnBot]

    stirrup = SteelDisposerStirrup(
        area=100,
        step=150,
        angle=90
    )
    section.stirrup = stirrup

    section_solver_input = ModelInputRcSecCheck(loads=loads, section=section)

    section_solver_input.criteria = [
        CheckableCriteriaEnum.SLE_NM,
        CheckableCriteriaEnum.SLE_F,
        CheckableCriteriaEnum.SLU_T,
        CheckableCriteriaEnum.SLU_NM
    ]
    section_solver_input.loadsInCriteria = [[0,1,2,3],[2,3],[4],[3,5,6,7]]

    solver = RcSecRectSolver()
    solver.setJobPath(str(tmp_path))
    solver.setModelInput(section_solver_input)
    solver.run(SolverOptions.STATIC)
    solver.reportName = "test_rc_solver_cl_usability_002"
    solver.buildReport(opt=ReportOption.REPORT_FROM_RUN)

@pytest.mark.codeaster
@pytest.mark.parametrize("file_name", [f"test_0{n}_RcSecSolver" for n in range(5, 6)])
def test_rc_sec_solver_thermal(tmp_path: Path, file_name: str) -> None:
    solver = RcSecRectSolver()
    # solver.setJobPath(str(Path(__file__).parent))
    solver.setJobPath(str(tmp_path))

    jsonObjectStatic = read_json(f"{file_name}.json")
    shutil.copyfile(src=Path(__file__).parent.joinpath(f"{file_name}.json"), dst=tmp_path.joinpath(f"{file_name}.json"))
    iDataStatic = ModelInputRcSecCheck(**jsonObjectStatic)

    # TODO: skip this if server is unavailable
    if CheckableCriteriaEnum.SLU_NM_FIRE in iDataStatic.criteria:
        jsonObject = read_json(f"{file_name}_Thermal.json")
        iDataThermal = ThermalMapSolverIn(**jsonObject)
        solver.setModelInput(iDataThermal)
        assert solver.run(SolverOptions.THERMAL, jobToken="pippo", test=True)

    solver.setModelInput(iDataStatic)
    solver.run(SolverOptions.STATIC)

    oData = solver.getModelOutput()
    assert isinstance(oData, ModelOutputRcSecCheck)
    for _i, c in enumerate(oData.results.resultsForCriteria):
        c.media = None

    # ------------------------------------------------
    # using this for generate not regression json file
    # ------------------------------------------------
    # outfile = Path(__file__).parent.joinpath(f"{file_name}_nr.json")
    # outfile.write_text(json.dumps(oData.model_dump(), indent=4))

    # Testing data persistence on file for oData
    data = ModelOutputRcSecCheck(**read_json(f"{file_name}_nr.json"))
    for _i, c in enumerate(data.results.resultsForCriteria):
        c.media = None

    assert data == oData

    solver.modelInputFile = f"{file_name}.json"
    shutil.copyfile(src=Path(__file__).parent.joinpath(f"{file_name}_nr.json"), dst=tmp_path.joinpath(f"{file_name}_nr.json"))
    solver.modelOutputFile = f"{file_name}_nr.json"
    solver.reportName = file_name
    solver.setLogLevel(3)
    assert solver.buildReport()

if __name__ == "__main__":
    sys.exit(pytest.main(Path(__file__)))



"""A package with very important objects in structural engineering domain.

The package is the most important of *pycivil* and contains a part of the
codebase in structural engineering domain. Starting from it, we can build
tools like post-processors (for Strand7, Midas etc.), checkers (for
standard codes like Eurocodes, Italian Codes).

"""


import json
from pathlib import Path
import pytest

from pycivil.EXAStructural.rcrecsolver.srvRcSecCheck import (
    ModelInputRcSecCheck,
    ModelOutputRcSecCheck,
    ReportBuilder,
)
from pycivil.EXAUtils.report import (
    Reporter,
    ReportTemplateEnum,
    getTemplatesPath,
)


@pytest.mark.needLatex
def test_report_latex_rcsec_check_01(tmp_path: Path):

    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_in_01.json"
    ) as jsonFile:
        jsonObjectIn = json.load(jsonFile)
        jsonFile.close()

    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_out_01.json"
    ) as jsonFile:
        jsonObjectOut = json.load(jsonFile)
        jsonFile.close()

    iData = ModelInputRcSecCheck(**jsonObjectIn)
    oData = ModelOutputRcSecCheck(**jsonObjectOut)

    rb = ReportBuilder(iData, oData)

    f = rb.buildFragment()

    latexTemplatePath = getTemplatesPath()
    reporter = Reporter(latexTemplatePath)

    reporter.linkFragments(template=ReportTemplateEnum.TEX_ENG_CAL, fragments=[f])
    reporter.compileDocument(path=str(tmp_path))
    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("report.pdf").exists()
    assert tmp_path.joinpath("report.tex").exists()


@pytest.mark.needLatex
def test_report_latex_rcsec_check_02(tmp_path: Path):

    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_in_02.json"
    ) as jsonFile:
        jsonObjectIn = json.load(jsonFile)
        jsonFile.close()

    with open(
        Path(__file__).parent / "test_report_latex_rcsec_check_out_02.json"
    ) as jsonFile:
        jsonObjectOut = json.load(jsonFile)
        jsonFile.close()

    iData = ModelInputRcSecCheck(**jsonObjectIn)
    oData = ModelOutputRcSecCheck(**jsonObjectOut)

    rb = ReportBuilder(iData, oData)

    f = rb.buildFragment()

    latexTemplatePath = getTemplatesPath()
    reporter = Reporter(latexTemplatePath)

    reporter.linkFragments(template=ReportTemplateEnum.TEX_ENG_CAL, fragments=[f])
    reporter.compileDocument(path=str(tmp_path))
    print(f"Temporary path for test is {tmp_path}")
    assert tmp_path.joinpath("report.pdf").exists()
    assert tmp_path.joinpath("report.tex").exists()



import json
from pathlib import Path
import pytest

from pycivil.EXAStructural.rcrecsolver.srvRcSecCheck import ModelInputRcSecCheck
from pycivil.EXAUtils.latexReportMakers import ConcreteSectionFB
from pycivil.EXAUtils.report import (
    Reporter,
    ReportTemplateEnum,
    getTemplatesPath,
)


@pytest.mark.needLatex
def test_report_latex_section_01(tmp_path: Path):

    with open(Path(__file__).parent / "test_report_latex_sections_01.json") as jsonFile:
        jsonObject = json.load(jsonFile)
        jsonFile.close()

    iData = ModelInputRcSecCheck(**jsonObject)

    print(json.dumps(iData.model_dump(), indent=3))

    section = iData.buildRcsRectangular()
    if section is not None:
        latexTemplatePath = getTemplatesPath()
        f = ConcreteSectionFB(latexTemplatePath, rcs=section).buildFragment()
        print(f.frags())
        reporter = Reporter(latexTemplatePath)
        reporter.linkFragments(
            template=ReportTemplateEnum.TEX_ENG_CAL, fragments=[f]
        )
        reporter.compileDocument(path=str(tmp_path))
        print(f"Temporary path for test is {tmp_path}")
        assert tmp_path.joinpath("report.pdf").exists()
        assert tmp_path.joinpath("report.tex").exists()


@pytest.mark.needLatex
def test_report_latex_section_02(tmp_path: Path):

    with open(Path(__file__).parent / "test_report_latex_sections_02.json") as jsonFile:
        jsonObject = json.load(jsonFile)
        jsonFile.close()

    iData = ModelInputRcSecCheck(**jsonObject)

    print(json.dumps(iData.model_dump(), indent=3))

    section = iData.buildRcsRectangular()
    if section is not None:
        latexTemplatePath = getTemplatesPath()
        f = ConcreteSectionFB(latexTemplatePath, rcs=section).buildFragment()
        print(f.frags())
        reporter = Reporter(latexTemplatePath)
        reporter.linkFragments(
            template=ReportTemplateEnum.TEX_ENG_CAL, fragments=[f]
        )
        reporter.compileDocument(path=str(tmp_path))
        print(f"Temporary path for test is {tmp_path}")
        assert tmp_path.joinpath("report.pdf").exists()
        assert tmp_path.joinpath("report.tex").exists()



import unittest
from pathlib import Path

from pycivil.EXAGeometry.geometry import Point2d
from pycivil.EXAStructural.modeler import SectionModeler
from pycivil.EXAStructural.plot import Geometry2DPlot, SectionPlot


class Test(unittest.TestCase):
    def test_101_modeler(self):
        # Build modeler
        md = SectionModeler()
        md.setLogLevel(0)

        # Add section and make it current
        md.addSection(1, True)

        # Add node for section
        md.addNode(1, 0, 0)
        md.addNode(2, 300, 0)
        md.addNode(3, 300, 600)
        md.addNode(4, 0, 600)

        md.addTriangle(1, 1, 2, 3)
        md.addTriangle(2, 3, 4, 1)

        # Add node and rebar on bottom
        md.addNode(20, 60, 40)
        md.addNode(21, 120, 40)
        md.addNode(22, 180, 40)
        md.addNode(23, 240, 40)

        diameterBot = 28
        md.addCircle(20, 20, diameterBot / 2)
        md.addCircle(21, 21, diameterBot / 2)
        md.addCircle(22, 22, diameterBot / 2)
        md.addCircle(23, 23, diameterBot / 2)

        # Add node and rebar on top
        md.addNode(10, 60, 560)
        md.addNode(11, 120, 560)
        md.addNode(12, 180, 560)
        md.addNode(13, 240, 560)

        diameterTop = 8
        md.addCircle(10, 10, diameterTop / 2)
        md.addCircle(11, 11, diameterTop / 2)
        md.addCircle(12, 12, diameterTop / 2)
        md.addCircle(13, 13, diameterTop / 2)

        # md.rotate(90, Point2d(150, 300))

        sp = SectionPlot()
        sp.modeler = md
        # sp.setView(SectionPlotView_Enum.SP_VIEW_MESH)

        self.assertTrue(md.buildDomain(15, 30))
        sp.plot()
        sp.save(Path(__file__).parent / "test_101_modeler_section.png")
        bbox = md.domainBounding()

        print(f"bounding Fz: ({bbox[0]},{bbox[1]})")
        print(f"bounding Mx: ({bbox[2]},{bbox[3]})")
        print(f"bounding My: ({bbox[4]},{bbox[5]})")
        # My = const -> (N,Mx)
        intersectionsMy = md.intersectAtMy(0, 1e-3, 1e-6)
        # N = const  -> (Mx, My)
        intersectionsN = md.intersectAtN(0, 1e-6, 1e-6)
        # Mx = const -> (N,My)
        intersectionsMx = md.intersectAtMx(0, 1e-3, 1e-6)

        print(f"Intersections vec lenght {len(intersectionsMy)}")

        gpMy = Geometry2DPlot(figsize=(7, 7))
        gpMy.addPointCloud(intersectionsMy)
        gpMy.showItems(True, False, True)
        gpMy.setTitles(
            [
                "Domain N-Mx in KNm",
            ]
        )
        gpMy.setXLabel(
            [
                "N [KN]",
            ]
        )
        gpMy.setYLabel(
            [
                "Mx [KNm]",
            ]
        )
        gpMy.plot()
        gpMy.save(Path(__file__).parent / "test_101_modeler_domain_01.png")

        gpN = Geometry2DPlot(figsize=(7, 7))
        gpN.addPointCloud(intersectionsN)
        gpN.showItems(False, False, True)
        gpN.setTitles(
            [
                "Domain Mx-My in KNm",
            ]
        )
        gpN.plot()
        gpN.save(Path(__file__).parent / "test_101_modeler_domain_02.png")

        gpMx = Geometry2DPlot(figsize=(7, 7))
        gpMx.addPointCloud(intersectionsMx)
        gpMx.showItems(False, False, True)
        gpMx.setTitles(
            [
                "Domain N,My in KNm",
            ]
        )
        gpMx.plot()
        gpMx.save(Path(__file__).parent / "test_101_modeler_domain_03.png")

        gpArray = Geometry2DPlot(3, 1, figsize=(7, 14))
        gpArray.addPointCloud(intersectionsMy)
        gpArray.addPointCloud(intersectionsN)
        gpArray.addPointCloud(intersectionsMx)
        pointArrayMy = [Point2d(0, 0), Point2d(100, -100)]
        pointArrayN = [Point2d(0, 0), Point2d(10, 100)]
        pointArrayMx = [Point2d(0, 0), Point2d(100, 200)]
        gpArray.addPointArray(pointArrayMy, ["green", "green"])
        gpArray.addPointArray(pointArrayN, ["green", "green"])
        gpArray.addPointArray(pointArrayMx, ["green", "red"])
        gpArray.showItems(False, False, True)
        gpArray.setTitles(
            [
                "Domain N-Mx in KNm",
                "Domain Mx-My in KNm",
                "Domain N,My in KNm",
            ]
        )
        gpArray.setXLabel(
            [
                "N [KN]",
                "Mx [KNm]",
                "N [KNm]",
            ]
        )
        gpArray.setYLabel(
            [
                "Mx [KN]",
                "My [KNm]",
                "My [KNm]",
            ]
        )
        gpArray.plot()
        gpArray.save(Path(__file__).parent / "test_101_modeler_domain_04.png")
        md.setLogLevel(1)
        md.printInfo()

    def test_102_modeler(self):
        # Build modeler
        md = SectionModeler()
        md.setLogLevel(0)

        # Add section and make it current
        md.addSection(1, True)

        # Add node for section
        md.addNode(1, 0, 0)
        md.addNode(2, 600, 0)
        md.addNode(3, 600, 300)
        md.addNode(4, 0, 300)

        md.addTriangle(1, 1, 2, 3)
        md.addTriangle(2, 3, 4, 1)

        # Add node and rebar on bottom
        md.addNode(20, 560, 60)
        md.addNode(21, 560, 120)
        md.addNode(22, 560, 180)
        md.addNode(23, 560, 240)

        diameterBot = 28
        md.addCircle(20, 20, diameterBot / 2)
        md.addCircle(21, 21, diameterBot / 2)
        md.addCircle(22, 22, diameterBot / 2)
        md.addCircle(23, 23, diameterBot / 2)

        # Add node and rebar on top
        md.addNode(10, 40, 60)
        md.addNode(11, 40, 120)
        md.addNode(12, 40, 180)
        md.addNode(13, 40, 240)

        diameterTop = 16
        md.addCircle(10, 10, diameterTop / 2)
        md.addCircle(11, 11, diameterTop / 2)
        md.addCircle(12, 12, diameterTop / 2)
        md.addCircle(13, 13, diameterTop / 2)

        sp = SectionPlot()
        sp.modeler = md
        # sp.setView(ViewEnum.SP_VIEW_TRIANGLES)
        sp.plot()
        sp.save(Path(__file__).parent / "test_102_modeler_section.png")

        self.assertTrue(md.buildDomain(15, 30))

        bbox = md.domainBounding()
        print(f"bounding Fz: ({bbox[0]},{bbox[1]})")
        print(f"bounding Mx: ({bbox[2]},{bbox[3]})")
        print(f"bounding My: ({bbox[4]},{bbox[5]})")
        # My = const -> (N,Mx)
        intersectionsMy = md.intersectAtMy(0, 1e-3, 1e-6)
        # N = const  -> (Mx, My)
        intersectionsN = md.intersectAtN(0, 1e-6, 1e-6)
        # Mx = const -> (N,My)
        intersectionsMx = md.intersectAtMx(0, 1e-3, 1e-6)

        print(f"Intersections vec lenght {len(intersectionsMy)}")

        gpMy = Geometry2DPlot(figsize=(7, 7))
        gpMy.addPointCloud(intersectionsMy)
        gpMy.showItems(False, False, True)
        gpMy.setTitles(
            [
                "Domain N-Mx in KNm",
            ]
        )
        gpMy.setXLabel(
            [
                "N [KN]",
            ]
        )
        gpMy.setYLabel(
            [
                "Mx [KNm]",
            ]
        )
        gpMy.plot()
        gpMy.save(Path(__file__).parent / "test_102_modeler_domain_01.png")

        gpN = Geometry2DPlot(figsize=(7, 7))
        gpN.addPointCloud(intersectionsN)
        gpN.showItems(False, False, True)
        gpN.setTitles(
            [
                "Domain Mx-My in KNm",
            ]
        )
        gpN.plot()
        gpN.save(Path(__file__).parent / "test_102_modeler_domain_02.png")

        gpMx = Geometry2DPlot(figsize=(7, 7))
        gpMx.addPointCloud(intersectionsMx)
        gpMx.showItems(False, False, True)
        gpMx.setTitles(
            [
                "Domain N,My in KNm",
            ]
        )
        gpMx.plot()
        gpMx.save(Path(__file__).parent / "test_102_modeler_domain_03.png")

        gpArray = Geometry2DPlot(3, 1, figsize=(7, 14))
        gpArray.addPointCloud(intersectionsMy)
        gpArray.addPointCloud(intersectionsN)
        gpArray.addPointCloud(intersectionsMx)
        pointArrayMy = [Point2d(0, 0), Point2d(100, -100)]
        pointArrayN = [Point2d(0, 0), Point2d(10, 100)]
        pointArrayMx = [Point2d(0, 0), Point2d(100, 200)]
        gpArray.addPointArray(pointArrayMy, ["green", "green"])
        gpArray.addPointArray(pointArrayN, ["green", "green"])
        gpArray.addPointArray(pointArrayMx, ["green", "red"])
        gpArray.showItems(False, False, True)
        gpArray.setTitles(
            [
                "Domain N-Mx in KNm",
                "Domain Mx-My in KNm",
                "Domain N,My in KNm",
            ]
        )
        gpArray.setXLabel(
            [
                "N [KN]",
                "Mx [KNm]",
                "N [KNm]",
            ]
        )
        gpArray.setYLabel(
            [
                "Mx [KN]",
                "My [KNm]",
                "My [KNm]",
            ]
        )
        gpArray.plot()
        gpArray.save(Path(__file__).parent / "test_102_modeler_domain_04.png")
        md.setLogLevel(1)
        md.printInfo()


if __name__ == "__main__":
    unittest.main()



from pathlib import Path
import getpass

from uuid import uuid4

from typing import List, Literal

from pycivil.EXAStructural.loads import (
    ForcesOnSection as Force,
    Frequency_Enum as Frequency,
    LimiteState_Enum as LimitState
)

from pycivil.EXAStructural.checkable import CheckableCriteriaEnum as TpCheck
from pycivil.EXAStructural.materials import (
    Concrete, ConcreteModel, ConcreteSteel, SteelModel
)

from pycivil.EXAStructural.sections import (
    SteelDisposerOnLine,
    SteelDisposerStirrup
)

from pycivil.EXAUtils.models import User, Project
from pycivil.EXAStructural.sections import RectangularShape, ShapePositionEnum, ConcreteSectionModel
from pycivil.EXAStructural.codes import Code

from pycivil.EXAStructural.rcrecsolver.srvRcSecCheck import (
    ModelInputRcSecCheck,
    RcSecRectSolver,
    SolverOptions,
    ReportOption
)

from pycivil.EXAUtils.logging import log

class RCRectCalculator:
    def __init__(self, project_name: str, section_name: str):
        self.__idf: List[int] = []
        self.__forces: List[Force] = []
        self.__mat_concrete: Concrete | None = None
        self.__mat_rebars: ConcreteSteel | None = None
        self.__rebars_disposer_on_line: List[SteelDisposerOnLine] = []
        self.__rebars_stirrup_single: SteelDisposerStirrup | None = None
        self.__shape: RectangularShape | None = None
        self.__criteria: List[TpCheck] = \
            [TpCheck.SLE_NM, TpCheck.SLE_F, TpCheck.SLU_T, TpCheck.SLU_NM]
        self.__loads_in_criteria: List[List[int]] = [[],[],[],[]]
        self.__jobPath: Path | None = None
        self.__report_name: str = "RCRectCalculator_Report"
        self.__solver: RcSecRectSolver = RcSecRectSolver()
        self.__solver_exit: bool | None = None
        self.__ll: Literal[0, 1, 2, 3] = 3
        self.__section_descr: str = section_name
        self.__project_brief: str = project_name

    def setDescription(self, descr: str) -> None:
        self.__section_descr = descr

    def setProjectBrief(self, brief: str) -> None:
        self.__project_brief = brief

    def setLogLevel(self, ll: Literal[0, 1, 2, 3]):
        self.__ll = ll
        self.__solver.setLogLevel(ll)

    def setMaterialConcrete(self,
                            code: Literal["NTC2008", "NTC2018", "EC2", "EC2:ITA", "NTC2018:RFI"],
                            code_str: str,
                            environment: Literal["not aggressive", "low-aggressive", "hight-aggressive"],
                            descr: str=""):
        self.__mat_concrete = Concrete()
        self.__mat_concrete.setMatDescr(descr)
        self.__mat_concrete.setByCode(Code(codeStr=code), code_str)
        self.__mat_concrete.setEnvironment(environment)

    def setMaterialRebars(self,
                          code: Literal["NTC2008", "NTC2018", "EC2", "EC2:ITA", "NTC2018:RFI"],
                          code_str: str,
                          environment: Literal["not sensitive", "sensitive"],
                          descr: str=""):
        self.__mat_rebars = ConcreteSteel()
        self.__mat_rebars.setMatDescr(descr)
        self.__mat_rebars.setByCode(Code(codeStr=code), code_str)
        self.__mat_rebars.setSensitivity(environment)

    def addForce(self,
                 N: float, M: float, T: float,
                 limit_state: Literal["ultimate", "serviceability", "accidental"],
                 check_required: List[Literal["SLE-NM", "SLE-F", "SLU-T", "SLU-NM"]],
                 frequency: Literal["quasi-permanent", "characteristic", "frequent", "ND"] = "ND",
                 descr: str=""):
        self.__forces.append(Force(
            Fx=N, My=M, Fz=T,
            descr=descr, id=len(self.__forces),
            limitState=LimitState(limit_state),
            frequency=Frequency(frequency)
        ))
        for tp in check_required:
            self.__loads_in_criteria[self.__criteria.index(TpCheck(tp))].append(len(self.__forces)-1)

    def addRebarsFromTop(self, num: float, diam: float,  dist_from_top: float, dist_rebars: float):
        self.__rebars_disposer_on_line.append(
            SteelDisposerOnLine(
                fromPos=ShapePositionEnum.MT,
                diameter=diam,
                steelInterDistance=dist_rebars,
                distanceFromPos=dist_from_top,
                number=num)
        )

    def addRebarsFromBot(self, num: float, diam: float,  dist_from_bot: float, dist_rebars: float):
        self.__rebars_disposer_on_line.append(
            SteelDisposerOnLine(
                fromPos=ShapePositionEnum.MB,
                diameter=diam,
                steelInterDistance=dist_rebars,
                distanceFromPos=dist_from_bot,
                number=num)
        )

    def setStirrup(self, area: float, step: float, angle):
        self.__rebars_stirrup_single = SteelDisposerStirrup()
        self.__rebars_stirrup_single.area = area
        self.__rebars_stirrup_single.step = step
        self.__rebars_stirrup_single.angle = angle

    def setDimensions(self, w: float, h: float, ids: int=-1, descr: str=""):
        self.__shape = RectangularShape(width=w, height=h, id=ids, descr=descr)

    def setJobPath(self, folder_path: str) -> bool:
        if not Path(folder_path).is_dir():
            log("ERR",f"Folder path {folder_path} is not a dir !!!", self.__ll)
            return False
        if not Path(folder_path).exists():
            log("ERR",f"Folder path {folder_path} doesn't exists !!!", self.__ll)
            return False
        log("INF", f"Job path {folder_path} successfully set", self.__ll)
        self.__jobPath = Path(folder_path)
        self.__solver.setJobPath(str(self.__jobPath))
        return True

    def run(self) -> bool:
        if self.__mat_concrete is None:
            log("ERR",f"Need to set concrete material with setMaterialConcrete() function !!!", self.__ll)
            return False
        concreteModel = ConcreteModel()
        concreteModel.fromMaterial(self.__mat_concrete)

        if self.__mat_rebars is None:
            log("ERR",f"Need to set rebars material with setMaterialRebars() function !!!", self.__ll)
            return False
        steelModel = SteelModel()
        steelModel.fromMaterial(self.__mat_rebars)

        if self.__shape is None:
            log("ERR",f"Need to set dimensions for concrete with setDimensions() function !!!", self.__ll)
            return False

        section = ConcreteSectionModel(
            concreteMat=concreteModel,
            steelMat=steelModel,
            shape=self.__shape,
            descr=self.__section_descr
        )

        if len(self.__rebars_disposer_on_line) == 0:
            if self.__shape is None:
                log("ERR", f"Need to add rebars with addRebarsFromTop or  addRebarsFromBot function !!!")
                return False
        section.disposerOnLine = self.__rebars_disposer_on_line

        if self.__rebars_stirrup_single is None:
            log("ERR",f"Need to add stirrup with setStirrup() function !!!", self.__ll)
            return False
        section.stirrup = self.__rebars_stirrup_single

        if len(self.__forces) == 0:
            log("ERR",f"Need to add forces with addForce() function !!!", self.__ll)
            return False

        section_solver_input = ModelInputRcSecCheck(loads=self.__forces, section=section)
        section_solver_input.criteria = self.__criteria
        section_solver_input.loadsInCriteria = self.__loads_in_criteria

        userName = ""
        try:
            userName = getpass.getuser()
        except KeyError as e:
            log("WRN",f"In this system user doesn't have name !!!", self.__ll)

        section_solver_input.user = User(usr=userName)
        section_solver_input.project = Project(uuid=uuid4(), brief=self.__project_brief)

        self.__solver.setModelInput(section_solver_input)
        self.__solver_exit = self.__solver.run(SolverOptions.STATIC)
        if self.__solver_exit:
            log("INF",f"Solver run with success.", self.__ll)
        else:
            log("ERR", f"Solver run with error !!!", self.__ll)
        return self.__solver_exit

    def buildReport(self) -> bool:
        if self.__solver_exit is None:
            log("ERR",f"You must use run function before build report !!!", self.__ll)
            return False
        if not self.__solver_exit:
            log("ERR",f"Solver exited with error, can't build a report !!!", self.__ll)
            return False
        log("INF",f"Building report ...", self.__ll)
        self.__solver.reportName = self.__report_name
        exit_val = self.__solver.buildReport(opt=ReportOption.REPORT_FROM_RUN)
        if exit_val:
            log("INF", f"done.", self.__ll)
        else:
            log("ERR", f"... error in building report !!!", self.__ll)
        return exit_val






import math
import os
from enum import Enum
from typing import Any, Optional, Union, List, Literal, Dict, Tuple, Set

import gmsh
import numpy as np
from numpy import uint64, uint64

from gmsh import model as gmsh_model
from jinja2 import Environment, FileSystemLoader


from pycivil.EXAUtils import logging as log, EXAExceptions as Ex

from pycivil.settings import ServerSettings

thisPath = os.path.dirname(os.path.abspath(__file__))
sp = ServerSettings().midas_templates_path

file_loader = FileSystemLoader(searchpath=sp)
env = Environment(loader=file_loader)
template = env.get_template("template-box.mgt")



class Dim(Enum):
    DIM_0D = 0
    DIM_1D = 1
    DIM_2D = 2
    DIM_3D = 3


class Elem(Enum):
    ELEM_TRIA = 3
    ELEM_QUAD = 4


class FEModel:
    configData = {
        "LoadCaseType": [
            {"id": "U", "description": "User Defined Load"},
            {"id": "D", "description": "Dead Load"},
            {"id": "L", "description": "Live Load"},
            {"id": "R", "description": "Roof Live Load"},
            {"id": "W", "description": "Wind Load on Structure"},
            {"id": "E", "description": "Earthquake"},
            {"id": "T", "description": "Temperature"},
            {"id": "S", "description": "Snow Load"},
            {"id": "R", "description": "Rain Load"},
            {"id": "IL", "description": "Live Load Impact"},
            {"id": "EP", "description": "Earth Pressure"},
            {"id": "B", "description": "Buoyancy"},
            {"id": "WP", "description": "Ground Water Pressure"},
            {"id": "FP", "description": "Fluid Pressure"},
            {"id": "IP", "description": "Ice Pressure"},
            {"id": "WL", "description": "Wind Load on Live Load"},
            {"id": "BK", "description": "Longitudinal Force from Live Load"},
            {"id": "CF", "description": "Centrifugal Force"},
            {"id": "RS", "description": "Rib Shortening"},
            {"id": "SH", "description": "Shrinkage"},
            {"id": "CR", "description": "Creep"},
            {"id": "PS", "description": "Pre-stressing force"},
            {"id": "ER", "description": "Erection Load"},
            {"id": "CO", "description": "Collision Load"},
            {"id": "CS", "description": "Construction Stage Load"},
            {"id": "BL", "description": "Braking Load"},
            {"id": "STL", "description": "Settlement"},
            {"id": "WPL", "description": "Wave Pressure"},
            {"id": "CL", "description": "Crowd Load"},
            {"id": "TG", "description": "Temperature Gradient"},
            {"id": "SW", "description": "Self Weight"}
        ]
    }
    def __init__(self, descr: str="", modelName: str="default"):

        # GMSH init
        gmsh.initialize()

        # Using OCC as BREP for geometry
        self.__cad = gmsh_model.occ

        # Adding model and make current
        gmsh_model.add(modelName)
        gmsh_model.setCurrent(modelName)

        self.__models = [modelName]

        print(gmsh_model.list())

        self.__description = descr
        self.__loadCase = {}
        self.__loads = {}
        self.__loadCombinations: Dict[str, List[float]] = {}
        self.__nodeSupports = {}
        self.__nodeSprings = {}

        # [Load Case][Node Tag][Tp]
        self.__nodeTemperatures: Dict[str, Dict[int, Dict[str, float]]] = {}

        # [Load Case][Frame Tag][Tp]
        self.__frameTemperatures: Dict[str, Dict[int, Dict[str, Tuple[float, float]]]] = {}

        # Values that don't depend on geometry or mesh
        # [Load Case] = value
        self.__initialTemperatures: Dict[str, float] = {}
        self.__inertialAcceleration: Dict[str, Tuple[float, float, float]] = {}

        # Springs evenly distribuited in local direction
        #
        # Dict[id_frame, Tuple[LCDIR1, LCDIR2, ONLYCOMPRESSION]]
        self.__framesSprings: Dict[int, Tuple[float, float, bool]] = {}

        self.__sectionShapes = {
            1: {"type": "RECTANGULAR", "dimensions": [1.0, 0.3], "name": "default"}
        }
        self.__framesShapes = {}

        self.__materialLibrary = {}
        self.__framesMaterial = {}

        self.__logLevel: Literal[0,1,2,3] = 3

        # [group tree name, physical]
        self.__groupTree: Dict[str,str] = {}

        # Link between Frames and phisical Names
        self.__framesPhysicalNames: Dict[int, List[str]] = {}

    def __assignShapesToFrames(self):
        pass

    def __log(self, tp, msg):
        log.log(tp, msg, self.__logLevel)

    @staticmethod
    def show():
        gmsh.fltk.run()

    def addMaterialToLibrary(
            self,
            idMaterial: int,
            modulus: float, # Mpa
            poisson: float, # ...
            name: str,
            density: float = 0.0, # kg/mc
            thermal_expansion: float= 0.0, # 1/C
            conductivity: float = 0.0, # KJ/s/m/C
            specific_heat: float = 0.0, # KJ/kg/C
    ):
        self.__materialLibrary[idMaterial] = {
            "name": name,
            "modulus": modulus,
            "poisson": poisson,
            "density": density,
            "thermal_expansion": thermal_expansion,
            "conductivity": conductivity,
            "specific_heat": specific_heat
        }

    def addSectionShape(self, idShape: int, name: str, tp: str, dim: Optional[List[float]]=None):
        if dim is None:
            dim = []

        if idShape in self.__sectionShapes.keys():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0007", "idShape for shape must be unique", idShape
            )

        if not isinstance(idShape, int):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004", "idShape must be a int", type(idShape)
            )

        if not isinstance(tp, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004", "name must be a str", type(name)
            )

        if not isinstance(tp, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004", "tp must be a str", type(tp)
            )

        if not all(isinstance(n, float) for n in dim):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0009", "dim must be float list", type(dim)
            )
        else:
            np.array(dim, dtype=np.float32)

        tps = ["RECTANGULAR"]

        if tp not in tps:
            raise Ex.EXAExceptions("(EXAStructuralModel)-0004", "tp unknown", type(tp))

        if tp == "RECTANGULAR":

            if len(dim) != 2:
                raise Ex.EXAExceptions(
                    "(EXAStructuralModel)-0004",
                    "For *RECTANGULAR* dim must have a two dim array",
                    len(dim),
                )

        self.__sectionShapes[idShape] = {
            "type": "RECTANGULAR",
            "dimensions": dim,
            "name": name,
        }

    def sectionShapes(self):
        return self.__sectionShapes

    def sectionMaterials(self):
        return self.__materialLibrary

    def assignFrameSectionShape(self, tagFrame: uint64, idShape: int):
        if not (isinstance(tagFrame, int) or isinstance(tagFrame, np.uint64)):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004", "tagFrame must be a int", type(tagFrame)
            )

        if not isinstance(idShape, int):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004", "idShape must be a int", type(idShape)
            )

        if idShape not in self.__sectionShapes.keys():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004",
                "idShape unknown in sectionShapes table",
                idShape,
            )

        self.__framesShapes[tagFrame] = idShape

    def assignFrameMaterial(self, tagFrame, idMaterial):
        if not (isinstance(tagFrame, int) or isinstance(tagFrame, np.uint64)):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004", "tagFrame must be a int", type(tagFrame)
            )

        if not isinstance(idMaterial, int):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004", "idMaterial must be a int", type(idMaterial)
            )

        if idMaterial not in self.__materialLibrary.keys():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004",
                "idMaterial unknown in materialLibrary table",
                idMaterial,
            )

        self.__framesMaterial[tagFrame] = idMaterial

    # llp:
    # assignFrameWinklerSupport assign a special type of support evenly distribuited to frame.
    # The main difference with addMultiframeSprings is that in last function subgrade modulus
    # will be transformed in nodal springs
    def assignFrameWinklerSupport(self, tagFrame: int, winkler_supports: Tuple[float, float, bool]):
        self.__framesSprings[tagFrame] = winkler_supports

    def getNodeTemperatures(self):
        return self.__nodeTemperatures

    def getFrameTemperatures(self):
        return self.__frameTemperatures

    def getInitialTemperatures(self):
        return self.__initialTemperatures

    def getInertialAcceleration(self):
        return self.__inertialAcceleration

    def setInertialAcceleration(self, loadCase: str, acc: Tuple[float, float, float]):
        if not isinstance(loadCase, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0018", "loadCase arg must be a string", type(loadCase)
            )

        if loadCase not in self.__loadCase.keys():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0019", "loadCase case unknown", loadCase
            )

        if not isinstance(acc, tuple):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0021", "acc must be float tuple", type(acc)
            )

        if len(acc) != 3:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0021", "acc must have len = 3", len(acc)
            )
        self.__inertialAcceleration[loadCase] = acc

    def setInitialTemperature(self, loadCase: str, temp: float):
        if not isinstance(loadCase, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0018", "First arg must be a string", type(loadCase)
            )

        if loadCase not in self.__loadCase.keys():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0019", "Load case unknown", loadCase
            )

        if not isinstance(temp, float):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0021", "temp must be float list", type(temp)
            )

        self.__initialTemperatures[loadCase] = temp

    def assignNodeTemperature(
            self,
            loadCase: str,
            tagNode: int,
            temp: float,
            tpTemp: Literal['fixed', 'reference', 'initial'] = 'fixed',
    ):
        if not isinstance(loadCase, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0018", "First arg must be a string", type(loadCase)
            )

        if loadCase not in self.__loadCase.keys():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0019", "Load case unknown", loadCase
            )

        if tagNode not in self.nodesTags():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0020", "Nodes tag unknown", tagNode
            )

        if not isinstance(temp, float):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0021", "temp must be float list", type(temp)
            )

        if not isinstance(tpTemp, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0021", "tpTemp must be str", type(tpTemp)
            )

        if loadCase not in self.__nodeTemperatures:
            self.__nodeTemperatures[loadCase] = {}

        if tagNode not in self.__nodeTemperatures[loadCase]:
            self.__nodeTemperatures[loadCase][tagNode] = {}

        self.__nodeTemperatures[loadCase][tagNode][tpTemp] = temp

    def assignFrameTemperature(
            self,
            loadCase: str,
            tagFrame: int,
            temp: Tuple[float, float],
            tpTemp: Literal['gradient'] = 'gradient',
    ):
        if not isinstance(loadCase, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0018", "First arg must be a string", type(loadCase)
            )

        if loadCase not in self.__loadCase:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0019", "loadCase unknown", loadCase
            )

        if tagFrame not in self.framesTags():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0013", "tagFrame unknown", tagFrame
            )

        if not isinstance(tpTemp, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0014", "tpTemp arg must be a string", type(tpTemp)
            )

        if not  tpTemp in ['gradient']:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0014", f"tpTemp arg not in {['gradient']}", type(tpTemp)
            )

        if not isinstance(temp, tuple):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0014", "temp arg must be a tuple", type(temp)
            )

        if len(temp) != 2:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0014", "temp arg must have len=2", len(temp)
            )

        if loadCase not in self.__frameTemperatures:
            self.__frameTemperatures[loadCase] = {}

        if tagFrame not in self.__frameTemperatures[loadCase]:
            self.__frameTemperatures[loadCase][tagFrame] = {}

        self.__frameTemperatures[loadCase][tagFrame][tpTemp] = temp


    def assignMultiFrameWinklerSupport(self, winkler_supports: Tuple[float, float, bool], tagsMacro=None, tagsFrames=None):
        if tagsMacro is None:
            tagsMacro = []
        if tagsFrames is None:
            tagsFrames = []

        if len(tagsMacro) == 0 and len(tagsFrames) == 0:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004",
                "tagsMacro or tagsFrames must be null",
                len(tagsMacro),
            )

        if len(tagsMacro) != 0 and len(tagsFrames) != 0:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004",
                "tagsMacro or tagsFrames must be not null",
                len(tagsMacro),
            )

        if len(tagsMacro) != 0:
            framesTags = self.getFramesFromMacroTags()
            for tag in tagsMacro:
                for _i, t in enumerate(framesTags[tag]):
                    self.assignFrameWinklerSupport(t, winkler_supports)
        else:
            for _i, t in enumerate(tagsFrames):
                self.assignFrameWinklerSupport(t, winkler_supports)

    def frameSectionShape(self):
        return self.__framesShapes

    def frameMaterial(self):
        return self.__framesMaterial

    def assignMultiFrameSectionShape(self, idShape: int, tagsMacro: Optional[List[uint64]]=None, tagsFrames: None=None):
        if tagsMacro is None:
            tagsMacro = []
        if tagsFrames is None:
            tagsFrames = []

        if len(tagsMacro) == 0 and len(tagsFrames) == 0:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004",
                "tagsMacro or tagsFrames must be null",
                len(tagsMacro),
            )

        if len(tagsMacro) != 0 and len(tagsFrames) != 0:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004",
                "tagsMacro or tagsFrames must be not null",
                len(tagsMacro),
            )

        if len(tagsMacro) != 0:
            framesTags = self.getFramesFromMacroTags()
            for tag in tagsMacro:
                for _i, t in enumerate(framesTags[tag]):
                    self.assignFrameSectionShape(t, idShape)
        else:
            for _i, t in enumerate(tagsFrames):
                self.assignFrameSectionShape(t, idShape)


    def assignMultiFrameMaterial(self, idMaterial, tagsMacro=None, tagsFrames=None):
        if tagsMacro is None:
            tagsMacro = []
        if tagsFrames is None:
            tagsFrames = []

        if len(tagsMacro) == 0 and len(tagsFrames) == 0:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004",
                "tagsMacro or tagsFrames must be null",
                len(tagsMacro),
            )

        if len(tagsMacro) != 0 and len(tagsFrames) != 0:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004",
                "tagsMacro or tagsFrames must be not null",
                len(tagsMacro),
            )

        if len(tagsMacro) != 0:
            framesTags = self.getFramesFromMacroTags()
            for tag in tagsMacro:
                for _i, t in enumerate(framesTags[tag]):
                    self.assignFrameMaterial(t, idMaterial)
        else:
            for _i, t in enumerate(tagsFrames):
                self.assignFrameMaterial(t, idMaterial)


    @staticmethod
    def clear():
        gmsh.clear()
        gmsh.finalize()

    @staticmethod
    def nodesTags() -> List[uint64]:
        nt = []
        entities = gmsh_model.getEntities()
        for e in entities:
            # Dimension and tag of the entity:
            dim = e[0]
            tag = e[1]
            # Get the mesh nodes for the entity (dim, tag):
            nodeTags, nodeCoords, nodeParams = gmsh.model.mesh.getNodes(dim, tag)
            nt += list(nodeTags)
        return nt


    @staticmethod
    def getFramesFromMacroTags(macroTags: List[Union[int,uint64]] | None=None) -> (
            Dict[int, Union[int,uint64]] | List[Union[int,uint64]]):
        """
        Retrive a list of mesh elements tags starting from macro element tags

        Args:
            macroTags ():

        Returns:
            If macroTags are given return a Dict with keys that are macroTags and values are list of mesh tags. In
            macroTags are None return a list of all mesh tags
        """
        if macroTags is None:
            macroTags = []

        ft = {}
        entities = gmsh_model.getEntities(1)
        for e in entities:
            # Dimension and tag of the entity:
            dim = e[0]
            tag = e[1]
            # Get the mesh nodes for the entity (dim, tag):
            elemTypes, elemTags, elemNodeTags = gmsh.model.mesh.getElements(dim, tag)
            ft[tag] = list(elemTags[0])
        if len(macroTags) == 0:
            return ft
        else:
            tags = []
            for i in macroTags:
                tags += ft[i]
            return tags


    def framesSupports(self):
        return self.__framesSprings

    @staticmethod
    def framesConnectivity():
        """
        Return a tag frame dictionary with key = frame tag and list of 2d tuple with node tags

        Returns:

        """
        elementTypes, elementTags, nodeTags = gmsh_model.mesh.getElements(1)
        connectivity = {}
        for i, tag in enumerate(elementTags[0]):
            connectivity[tag] = (nodeTags[0][i*2], nodeTags[0][i*2+1])
        return connectivity

    @staticmethod
    def nodeCoords(nodeTags: int):
        coord, _, _, _ = gmsh_model.mesh.getNode(nodeTags)
        return coord

    @staticmethod
    def framesTags() -> List[uint64]:
        ft = []
        entities = gmsh_model.getEntities(1)
        for e in entities:
            # Dimension and tag of the entity:
            dim = e[0]
            tag = e[1]
            # Get the mesh nodes for the entity (dim, tag):
            elemTypes, elemTags, elemNodeTags = gmsh.model.mesh.getElements(dim, tag)
            ft += list(elemTags[0])
        return ft

    @staticmethod
    def getFrameNodesFromMacro(macroTags: List[List[int|uint64]] | List[int] | int | None = None, unique: bool = False) -> (
            Dict[int, List[Tuple[int, int]]] | List[Tuple[int, int]] | Set[int]):
        """
        Get nodes related to macro-elements

        Args:
            unique (): if True and MacroTags is given get a set o unique nodes tags. Default is False
            macroTags (): if macroTags is None get a dictionary with {macroTag_i: [node1_el_1, node2_el_1], ... ,
                [node1_el_n, node2_el_n]}. If macroTags is a list of macro elements get a list of nodes as
                [node1_el_1, node2_el_1], ... , [node1_el_n, node2_el_n]

        Returns:
            ():
        """
        if macroTags is None:
            macroTags = []
        else:
            if isinstance(macroTags, list):
                if len(macroTags) == 0:
                    return []
                else:
                    if isinstance(macroTags[0], list):
                        u_lst = []
                        for m in macroTags:
                            u_lst.extend(m)
                        macroTags = u_lst
            else:
                macroTags = [macroTags]

        ft = {}
        entities = gmsh_model.getEntities(1)
        unique_tags = set()
        for e in entities:
            # Dimension and tag of the entity:
            dim = e[0]
            tag = e[1]
            # Get the mesh nodes for the entity (dim, tag):
            elemTypes, elemTags, elemNodeTags = gmsh.model.mesh.getElements(dim, tag)
            nt = []
            for a in elemNodeTags[0].reshape(round(len(elemNodeTags[0]) / 2), 2):
                nt += [list(a)]
            ft[tag] = nt

        if len(macroTags) == 0:
            return ft
        else:
            tags = []
            for i in macroTags:
                tags += ft[i]

            if unique:
                for n in tags:
                    unique_tags.add(n[0])
                    unique_tags.add(n[1])
                return list(unique_tags)

            return tags

    @staticmethod
    def framesNodeTags() -> List[List[uint64]]:
        ft = []
        entities = gmsh_model.getEntities(1)
        for e in entities:
            # Dimension and tag of the entity:
            dim = e[0]
            tag = e[1]
            # Get the mesh nodes for the entity (dim, tag):
            elemTypes, elemTags, elemNodeTags = gmsh.model.mesh.getElements(dim, tag)
            nt = []
            for a in elemNodeTags[0].reshape(round(len(elemNodeTags[0]) / 2), 2):
                nt += [list(a)]
            ft += nt

        return ft

    @staticmethod
    def nodesCoords() -> List[List[float]]:
        nc = []
        entities = gmsh_model.getEntities()
        for e in entities:
            # Dimension and tag of the entity:
            dim = e[0]
            tag = e[1]

            # Get the mesh nodes for the entity (dim, tag):
            nodeTags, nodeCoords, nodeParams = gmsh.model.mesh.getNodes(dim, tag)
            nc += nodeCoords.reshape(round(len(nodeCoords) / 3), 3).tolist()
        return nc


    @staticmethod
    def nodesGroups():
        ng = {}
        entities = gmsh_model.getEntities(0)
        for e in entities:
            # Dimension and tag of the entity:
            dim = e[0]
            tag = e[1]
            ng[tag] = list(gmsh_model.getPhysicalGroupsForEntity(dim, tag))
        return ng


    @staticmethod
    def framesGroups():
        fg = {}
        entities = gmsh_model.getEntities(1)
        for e in entities:
            # Dimension and tag of the entity:
            dim = e[0]
            tag = e[1]
            fg[tag] = list(gmsh_model.getPhysicalGroupsForEntity(dim, tag))
        return fg


    @staticmethod
    def printNodes():
        print("** start Nodes **")
        entities = gmsh_model.getEntities()
        for e in entities:
            # Dimension and tag of the entity:
            dim = e[0]
            tag = e[1]

            # Get the mesh nodes for the entity (dim, tag):
            nodeTags, nodeCoords, nodeParams = gmsh.model.mesh.getNodes(dim, tag)
            print("nodeTags", nodeTags)
            print("nodeCoords", nodeCoords)
            print("phisicalGroup", gmsh_model.getPhysicalGroupsForEntity(dim, tag))
        print("** end Nodes **")


    @staticmethod
    def printFrames():
        print("** start Frames **")
        entities = gmsh_model.getEntities(1)
        for e in entities:
            # Dimension and tag of the entity:
            dim = e[0]
            tag = e[1]

            # Get the mesh elements for the entity (dim, tag):
            elemTypes, elemTags, elemNodeTags = gmsh.model.mesh.getElements(dim, tag)
            print("elemTypes", elemTypes)
            print("elemTags", elemTags)
            print("elemNodeTags", elemNodeTags)
            print("phisicalGroup", gmsh_model.getPhysicalGroupsForEntity(dim, tag))
        print("** end Frames **")


    @staticmethod
    def printElements():
        entities = gmsh_model.getEntities()
        for e in entities:
            # Dimension and tag of the entity:
            dim = e[0]
            tag = e[1]

            # Get the mesh elements for the entity (dim, tag):
            elemTypes, elemTags, elemNodeTags = gmsh.model.mesh.getElements(dim, tag)
            print("elemTypes", elemTypes)
            print("elemTags", elemTags)
            print("elemNodeTags", elemNodeTags)
            print("phisicalGroup", gmsh_model.getPhysicalGroupsForEntity(dim, tag))

    # llp: load combination list is related to load case. Order is the same. Starting from Python 3.6
    #      dictionaries hold ordee also. Then self.__loads even dict is ordered

    def addLoadCombination(self, keyCombination: str, combination: List[float]):
        self.__loadCombinations[keyCombination] = combination

    def getLoadCombinations(self):
        return self.__loadCombinations

    def addLoadCase(self, idCase: str, tp: str="U", descr: str=""):
        if not isinstance(idCase, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0002", "First arg must be a string", idCase
            )
        if not isinstance(tp, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0003", "Second arg must be a string", tp
            )
        else:
            find = False
            for v in self.configData["LoadCaseType"]:
                if v["id"] == tp:
                    find = True
                    print(
                        "Type *{}* of load case generic *{}*: {} ---> {}".format(
                            idCase, v["id"], v["description"], descr
                        )
                    )
                    break
            if not find:
                raise Ex.EXAExceptions("(EXAStructuralModel)-0005", "Type unknown", tp)

        if not isinstance(tp, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0004", "Third arg must be a string", descr
            )

        if idCase in self.__loadCase.keys():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0006", "Load case must be unique", idCase
            )

        self.__loadCase[idCase] = [tp, descr]

    def getLoadCases(self):
        return self.__loadCase

    def getLoads(self):
        return self.__loads

    def addSelfWeight(self, loadCase: str, GCS: List[float]):
        if not isinstance(loadCase, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0007", "First arg must be a string", loadCase
            )

        if loadCase not in self.__loadCase.keys():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0008", "Load case unknown", loadCase
            )

        if not all(isinstance(n, float) for n in GCS):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0009", "GCS must be float list", GCS
            )

        if len(GCS) != 3:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0010", "GCS must have len 3", len(GCS)
            )

        if loadCase not in self.__loads:
            self.__loads[loadCase] = {}

        self.__loads[loadCase]["selfWeight"] = GCS

    def addMultiFrameWinklerSpring(self, tagsMacro: List[uint64], tp: str, dirSprings: str, subgradeModulus: float, Bref: int):
        # subgradeModulus: modulo di sottofondo alla Winkler
        # Bref: largezza di riferimento per il calcolo della pressione

        if not isinstance(tagsMacro, list):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0011", "tagsMacro must be a list", type(tagsMacro)
            )

        nodesTags = self.nodesTags()
        nodesCoords = self.nodesCoords()
        framesTags = self.getFramesFromMacroTags()
        frameNodes = self.getFrameNodesFromMacro()

        for tag in tagsMacro:
            # TODO: move this in separate function for subFrames (meshed frames) and add tagsFrames
            #       in function args with two cases
            for i, _t in enumerate(framesTags[tag]):
                n1 = frameNodes[tag][i][0]
                n2 = frameNodes[tag][i][1]
                n1_coords = nodesCoords[nodesTags.index(n1)]
                n2_coords = nodesCoords[nodesTags.index(n2)]

                if "DX" in dirSprings:
                    idxDir1 = 1
                    idxDir2 = 2
                elif "DY" in dirSprings:
                    idxDir1 = 0
                    idxDir2 = 2
                else: # "DZ" in dir:
                    idxDir1 = 0
                    idxDir2 = 0

                tributaryLength = math.sqrt(
                    pow(n2_coords[idxDir1] - n1_coords[idxDir1], 2)
                    + pow(n2_coords[idxDir2] - n1_coords[idxDir2], 2)
                )

                stiffness = tributaryLength * subgradeModulus * Bref

                self.addNodeSpring(n1, tp, dirSprings, stiffness, add=True)
                self.addNodeSpring(n2, tp, dirSprings, stiffness, add=True)

    def addMultiFrameLoadHydro(
            self,
            loadCase: str,
            tagsMacro: List[int|uint64],
            dirLoad: Literal['GCX', 'GCY', 'GCZ'],
            gamma: float,
            K: float,
            H0: float,
            p0: float,
            Bref: float,
            local: bool=False,
            coordVariation: Literal['X', 'Y', 'Z'] = 'Z',
            oneSignPositive: bool | None = None,
            constValue: bool = False
    ):
        """
        Add a linear load on macro-element

        Args:
            constValue (): if true value is considered constant on frame p0 * Bref
            loadCase (): load case as str
            tagsMacro (): macro-element tag
            dirLoad (): load direction
            gamma (): peso del terreno specifico
            K (): coefficiente di spinta del terreno
            H0 (): distanza dallo zero della copertura del terreno
            p0 (): pressione di partenza ovvero il sovraccarico
            Bref (): largezza di riferimento per il calcolo della pressione
            local (): reference system for pressure direction
            coordVariation (): coordinate in global reference that for variation
            oneSignPositive (): If None function considers positive and negative values, If True function considers
                only positive values and negative values if it is False
        """
        if not isinstance(loadCase, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0011",
                "First arg must be a string",
                type(loadCase),
            )

        if not isinstance(tagsMacro, list):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0011", "tagsMacro must be a list", type(tagsMacro)
            )

        if dirLoad not in ("GCX", "GCY", "GCZ"):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0022",
                "Third arg *dir* must be GCX, GCY, GCZ",
                dirLoad,
            )

        nodesTags = self.nodesTags()
        nodesCoords = self.nodesCoords()
        framesTags = self.getFramesFromMacroTags()
        frameNodes = self.getFrameNodesFromMacro()

        if coordVariation == 'X':
            coords_idx = 0
        elif coordVariation == 'Y':
            coords_idx = 1
        else: # coordVariation == 'Z'
            coords_idx = 2

        for tag in tagsMacro:
            # TODO: move this in separate function for subFrames (meshed frames) and add tagsFrames
            #       in function args with two cases
            for i, t in enumerate(framesTags[tag]):
                n1 = frameNodes[tag][i][0]
                n2 = frameNodes[tag][i][1]
                n1_coords = nodesCoords[nodesTags.index(uint64(n1))]
                n2_coords = nodesCoords[nodesTags.index(uint64(n2))]
                press_1 = ((H0 - n1_coords[coords_idx]) * gamma * K + p0) * Bref
                press_2 = ((H0 - n2_coords[coords_idx]) * gamma * K + p0) * Bref
                rel_dist_from_I_node_I = 0.0
                rel_dist_from_I_node_J = 1.0
                if not constValue:
                    if oneSignPositive is not None and press_1 != 0.0 or press_2 != 0.0:
                        zero_rel_dist_from_I = abs(press_1 / (abs(press_1) + abs(press_2)))
                        if oneSignPositive:
                            if press_1 <= 0.0 and press_2 <= 0.0:
                                press_1 = 0.0
                                press_2 = 0.0
                            elif press_1 * press_2 < 0.0:
                                if press_1 < 0.0:
                                    press_1 = 0.0
                                    rel_dist_from_I_node_I = 1- zero_rel_dist_from_I
                                else: # press_2 < 0.0:
                                    press_2 = 0.0
                                    rel_dist_from_I_node_J = zero_rel_dist_from_I
                            else:
                                pass
                        else: # not oneSignPositive
                            if press_1 >= 0.0 and press_2 >= 0.0:
                                press_1 = 0.0
                                press_2 = 0.0
                            elif press_1 * press_2 < 0.0:
                                if press_1 > 0.0:
                                    press_1 = 0.0
                                    rel_dist_from_I_node_I = 1 - zero_rel_dist_from_I
                                else: # press_2 < 0.0:
                                    press_2 = 0.0
                                    rel_dist_from_I_node_J = zero_rel_dist_from_I
                            else:
                                pass

                else: # constValue == True
                    press_1 = p0 * Bref
                    press_2 = press_1
                    lenght = math.sqrt((n1_coords[coords_idx]-n2_coords[coords_idx])**2)
                    if n1_coords[coords_idx] >= H0 and n2_coords[coords_idx] >= H0:
                        rel_dist_from_I_node_I = 0.0
                        rel_dist_from_I_node_J = 1.0
                        press_1 = 0.0
                        press_2 = 0.0
                    else:
                        if n1_coords[coords_idx] <= H0 and n2_coords[coords_idx] <= H0:
                            rel_dist_from_I_node_I = 0.0
                            rel_dist_from_I_node_J = 1.0
                        else:
                            zero_rel_dist_from_I = abs(abs(H0) - abs(n1_coords[coords_idx])) / lenght
                            if n1_coords[coords_idx] < H0 <= n2_coords[coords_idx]:
                                rel_dist_from_I_node_I = 0.0
                                rel_dist_from_I_node_J = 1.0 - zero_rel_dist_from_I
                            else:
                                rel_dist_from_I_node_I = zero_rel_dist_from_I
                                rel_dist_from_I_node_J = 1.0


                if dirLoad == "GCX":
                    self.addFrameLoad(
                        loadCase, uint64(t), tp="force",
                        GCX1=[rel_dist_from_I_node_I, press_1], GCX2=[rel_dist_from_I_node_J, press_2], local=local
                    )
                elif dirLoad == "GCY":
                    self.addFrameLoad(
                        loadCase, uint64(t), tp="force",
                        GCY1=[rel_dist_from_I_node_I, press_1], GCY2=[rel_dist_from_I_node_J, press_2], local=local
                    )
                elif dirLoad == "GCZ":
                    self.addFrameLoad(
                        loadCase, uint64(t), tp="force",
                        GCZ1=[rel_dist_from_I_node_I, press_1], GCZ2=[rel_dist_from_I_node_J, press_2], local=local
                    )
                else:
                    raise Ex.EXAExceptions(
                        "(EXAStructuralModel)-0011", "unknown error in dirLoad", type(dirLoad)
                    )

    def addMultiFrameLoad(
            self, loadCase: str, tagsFrames: List[uint64], tp: str="force",
            GCX1: Optional[List[float]]=None, GCX2: Optional[List[float]]=None,
            GCY1: Optional[List[float]]=None, GCY2: Optional[List[float]]=None,
            GCZ1: Optional[List[float]]=None, GCZ2: Optional[List[float]]=None,
            local: bool=False
    ):
        if GCX1 is None:
            GCX1 = []
        if GCX2 is None:
            GCX2 = []
        if GCY1 is None:
            GCY1 = []
        if GCY2 is None:
            GCY2 = []
        if GCZ1 is None:
            GCZ1 = []
        if GCZ2 is None:
            GCZ2 = []

        if not isinstance(tagsFrames, list):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0011",
                "tagsFrames must be a list",
                type(tagsFrames),
            )

        for tag in tagsFrames:
            self.addFrameLoad(loadCase, tag, tp, GCX1, GCX2, GCY1, GCY2, GCZ1, GCZ2, local)

    def addFrameLoad(
            self, loadCase: str, tagFrame: uint64, tp: str="force",
            GCX1: Optional[List[float|Any]]=None, GCX2: Optional[List[Union[float, Any]]]=None,
            GCY1: Optional[List[float]]=None, GCY2: Optional[List[float]]=None,
            GCZ1: Optional[List[float]]=None, GCZ2: Optional[List[float]]=None,
            local: bool=False
    ):
        if GCX1 is None:
            GCX1 = []
        if GCX2 is None:
            GCX2 = []
        if GCY1 is None:
            GCY1 = []
        if GCY2 is None:
            GCY2 = []
        if GCZ1 is None:
            GCZ1 = []
        if GCZ2 is None:
            GCZ2 = []

        if not isinstance(loadCase, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0011",
                "First arg must be a string",
                type(loadCase),
            )

        if loadCase not in self.__loadCase.keys():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0012", "Load case unknown", loadCase
            )

        if tagFrame not in self.framesTags():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0013", "Frames tag unknown", tagFrame
            )

        if not isinstance(tp, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0014", "Third arg must be a string", type(tp)
            )

        if all(["force" != tp, "moment" != tp]):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0015", "tp option *force* or *moment*", tp
            )

        if not isinstance(GCX1, list) or \
           not isinstance(GCX2, list) or \
           not isinstance(GCY1, list) or \
           not isinstance(GCY2, list) or \
           not isinstance(GCZ1, list) or \
           not isinstance(GCZ2, list) :
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0016", "GCX1 ... GCZ2 arg must be lists", type(tp)
            )

        cond_null_X = all([len(GCX1) == 0, len(GCX2) == 0])
        cond_null_Y = all([len(GCY1) == 0, len(GCY2) == 0])
        cond_null_Z = all([len(GCZ1) == 0, len(GCZ2) == 0])

        if not cond_null_X:
            if len(GCX1) != 2:
                raise Ex.EXAExceptions(
                    "(EXAStructuralModel)-0016", "GCX1 must have len 2", len(GCX1)
                )

            if not all(isinstance(n, float) for n in GCX1):
                raise Ex.EXAExceptions(
                    "(EXAStructuralModel)-0017", "GCX1 must be float list", GCX1
                )

            if len(GCX2) != 2 and len(GCX2) != 0:
                raise Ex.EXAExceptions(
                    "(EXAStructuralModel)-0018", "GCX2 must have len 2 or 0", GCX2
                )

            if len(GCX2) == 0:
                GCX2 = GCX1[:]
                GCX2[0] = 1.0

        if not cond_null_Y:
            if len(GCY1) != 2:
                raise Ex.EXAExceptions(
                    "(EXAStructuralModel)-0019", "GCY1 must have len 2", len(GCY1)
                )

            if not all(isinstance(n, float) for n in GCY1):
                raise Ex.EXAExceptions(
                    "(EXAStructuralModel)-0017", "GCY1 must be float list", GCY1
                )

            if len(GCY2) != 2 and len(GCY2) != 0:
                raise Ex.EXAExceptions(
                    "(EXAStructuralModel)-0021", "GCX2 must have len 2 or 0", GCY2
                )

            if len(GCY2) == 0:
                GCY2 = GCY1[:]
                GCY2[0] = 1.0

        if not cond_null_Z:
            if len(GCZ1) != 2:
                raise Ex.EXAExceptions(
                    "(EXAStructuralModel)-0022", "GCZ1 must have len 2", len(GCZ1)
                )

            if not all(isinstance(n, float) for n in GCZ1):
                raise Ex.EXAExceptions(
                    "(EXAStructuralModel)-0017", "GCZ1 must be float list", GCZ1
                )

            if len(GCZ1) != 2 and len(GCZ1) != 0:
                raise Ex.EXAExceptions(
                    "(EXAStructuralModel)-0024", "GCX2 must have len 2 or 0", GCZ1
                )

            if len(GCZ2) == 0:
                GCZ2 = GCZ1[:]
                GCZ2[0] = 1.0

        if loadCase not in self.__loads:
            self.__loads[loadCase] = {}

        if "frameLoad" not in self.__loads[loadCase]:
            self.__loads[loadCase]["frameLoad"] = {}

        if tagFrame not in self.__loads[loadCase]["frameLoad"].keys():
            self.__loads[loadCase]["frameLoad"][tagFrame] = {}

        if tp not in self.__loads[loadCase]["frameLoad"][tagFrame].keys():
            self.__loads[loadCase]["frameLoad"][tagFrame][tp] = {}

        sfx = "G"
        if local:
            sfx = "L"

        if not cond_null_X:
            self.__loads[loadCase]["frameLoad"][tagFrame][tp][sfx+"CX1"] = GCX1
            self.__loads[loadCase]["frameLoad"][tagFrame][tp][sfx+"CX2"] = GCX2

        if not cond_null_Y:
            self.__loads[loadCase]["frameLoad"][tagFrame][tp][sfx+"CY1"] = GCY1
            self.__loads[loadCase]["frameLoad"][tagFrame][tp][sfx+"CY2"] = GCY2

        if not cond_null_Z:
            self.__loads[loadCase]["frameLoad"][tagFrame][tp][sfx+"CZ1"] = GCZ1
            self.__loads[loadCase]["frameLoad"][tagFrame][tp][sfx+"CZ2"] = GCZ2

    def addMultiNodeLoad(self, loadCase, tagsNodes, NCS):
        if not isinstance(tagsNodes, list):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0018", "tagsNodes arg must be a list", tagsNodes
            )

        for tag in tagsNodes:
            self.assignNodeLoad(loadCase, tag, NCS)

    def assignNodeLoad(self, loadCase, tagNode, NCS):
        if not isinstance(loadCase, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0018", "First arg must be a string", loadCase
            )

        if loadCase not in self.__loadCase.keys():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0019", "Load case unknown", loadCase
            )

        if tagNode not in self.nodesTags():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0020", "Nodes tag unknown", tagNode
            )

        if not all(isinstance(n, float) for n in NCS):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0021", "NCS must be float list", NCS
            )

        if len(NCS) != 6:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0022", "NCS must have len 6", len(NCS)
            )

        if loadCase not in self.__loads:
            self.__loads[loadCase] = {}

        if "nodeLoad" not in self.__loads[loadCase]:
            self.__loads[loadCase]["nodeLoad"] = {}

        self.__loads[loadCase]["nodeLoad"][tagNode] = {"NCS": NCS}

    def addMultiNodeContraints(self, tagsNodes, dof):
        if not isinstance(tagsNodes, list):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0018", "tagsNodes arg must be a list", tagsNodes
            )

        for tag in tagsNodes:
            self.addNodeContraints(tag, dof)

    def addNodeContraints(self, tagNode, dof):
        if tagNode not in self.nodesTags():
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0034", "Nodes tag unknown", tagNode
            )

        if not all(isinstance(n, int) for n in dof):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0021", "dof must be int list", dof
            )

        if len(dof) != 6:
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0022", "dof must have len 6", len(dof)
            )

        self.__nodeSupports[tagNode] = dof

    def getNodeContraints(self):
        return self.__nodeSupports

    def addMultiNodeSpring(self, tagsNodes, tp, dirSpring, stiffness):
        if not isinstance(tagsNodes, list):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0018", "tagsNodes arg must be a list", tagsNodes
            )

        for tag in tagsNodes:
            self.addNodeSpring(tag, tp, dirSpring, stiffness)

    def addNodeSpring(self, tagNode: uint64, tp: str, dirSpring: str, stiffness: float, add: bool=False):
        if not (isinstance(tagNode, int) or isinstance(tagNode, np.uint64)):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0022",
                "First arg *tagNode* must be int or numpy.uint64",
                tagNode,
            )

        if not isinstance(tp, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0022", "Second arg *tp* must be str", tp
            )

        if tp not in ("LINEAR", "COMP", "TENS"):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0022",
                "Second arg *tp* must be LINEAR, COMP or TENS",
                tp,
            )

        if not isinstance(dirSpring, str):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0022", "Third arg *dir* must be str", dirSpring
            )

        if dirSpring not in ("DX+", "DX-", "DY+", "DY-", "DZ+", "DZ-"):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0022", "Third arg *dir* must be DX+ ... DZ-", dirSpring
            )

        if not isinstance(stiffness, float):
            raise Ex.EXAExceptions(
                "(EXAStructuralModel)-0022",
                "Fourth arg *stiffness* must be float",
                stiffness,
            )

        if tagNode not in self.__nodeSprings.keys():
            self.__nodeSprings[tagNode] = {}

        if tp not in self.__nodeSprings[tagNode].keys():
            self.__nodeSprings[tagNode][tp] = {}

        if dirSpring in self.__nodeSprings[tagNode][tp] and add:
            self.__nodeSprings[tagNode][tp][dirSpring] += stiffness
        else:
            self.__nodeSprings[tagNode][tp][dirSpring] = stiffness

    def getNodeSprings(self):
        return self.__nodeSprings

    @staticmethod
    def getNodesPhysicalName():
        groups = gmsh_model.getPhysicalGroups(0)
        names = []
        for g in groups:
            names.append(gmsh_model.getPhysicalName(g[0], g[1]))
        return names

    @staticmethod
    def getPhisicalNames() -> List[str]:
        """
        Get unique physical names defined.
        If many groups exist with same name returns unique values

        Returns:
            List of unique group names
        """
        groups = gmsh_model.getPhysicalGroups()
        names = []
        for g in groups:
            names.append(gmsh_model.getPhysicalName(g[0], g[1]))
        return list(set(names))

    @staticmethod
    def getFramesPhysicalName():
        groups = gmsh_model.getPhysicalGroups(1)
        names = []
        for g in groups:
            names.append(gmsh_model.getPhysicalName(g[0], g[1]))
        return names

    @staticmethod
    def getNodesByPhysicalName(name):
        groups = gmsh_model.getPhysicalGroups(0)
        for g in groups:
            if name == gmsh_model.getPhysicalName(g[0], g[1]):
                return list(gmsh_model.getEntitiesForPhysicalGroup(g[0], g[1]))
        return []


    @staticmethod
    def getNodesByPhysicalGroup(tagGroup):
        groups = gmsh_model.getPhysicalGroups(0)
        for g in groups:
            if g[1] == tagGroup:
                return list(gmsh_model.getEntitiesForPhysicalGroup(g[0], g[1]))
        return []


    @staticmethod
    def getMacroFramesByPhysicalName(name: str) -> List[uint64]:
        groups = gmsh_model.getPhysicalGroups(1)
        for g in groups:
            if name == gmsh_model.getPhysicalName(g[0], g[1]):
                return list(gmsh_model.getEntitiesForPhysicalGroup(g[0], g[1]))
        return []


    @staticmethod
    def getFramesByPhysicalGroup(tagGroup):
        groups = gmsh_model.getPhysicalGroups(1)
        for g in groups:
            if g[1] == tagGroup:
                return list(gmsh_model.getEntitiesForPhysicalGroup(g[0], g[1]))
        return []

    def __save_mgt(self, fileName):
        # ----------------
        # *NODE    ; Nodes
        # ; iNO, X, Y, Z
        # ----------------
        nodes = []
        nodesTags = self.nodesTags()
        nodesCoords = self.nodesCoords()
        for i, n in enumerate(nodesTags):
            strFormat = "{id:d},{x:.6f},{y:.6f},{z:.6f}"
            formatted = strFormat.format(
                id=n, x=nodesCoords[i][0], y=nodesCoords[i][1], z=nodesCoords[i][2]
            )
            nodes.append(formatted)

        # ------------------------------------------------------------------------------------
        # *ELEMENT    ; Elements
        # ; iEL, TYPE, iMAT, iPRO, iN1, iN2, ANGLE, iSUB,                     ; Frame  Element
        # ; iEL, TYPE, iMAT, iPRO, iN1, iN2, ANGLE, iSUB, EXVAL, EXVAL2, bLMT ; Comp/Tens Truss
        # ; iEL, TYPE, iMAT, iPRO, iN1, iN2, iN3, iN4, iSUB, iWID , LCAXIS    ; Planar Element
        # ; iEL, TYPE, iMAT, iPRO, iN1, iN2, iN3, iN4, iN5, iN6, iN7, iN8     ; Solid  Element
        # ------------------------------------------------------------------------------------
        frames = []
        framesTags = self.framesTags()
        framesNodeTags = self.framesNodeTags()

        for i, f in enumerate(framesTags):
            if f in self.__framesShapes.keys():
                idShape = self.__framesShapes[f]
            else:
                idShape = 1
            strFormat = "{id:d},BEAM,1,{iPRO:d},{idStart:d},{idEnd:d},0,0"
            formatted = strFormat.format(
                id=f,
                iPRO=idShape,
                idStart=framesNodeTags[i][0],
                idEnd=framesNodeTags[i][1],
            )
            frames.append(formatted)

        # -----------------------------------------------------------------------------------------------------------
        # *SECTION    ; Section
        # ; iSEC, TYPE, SNAME, [OFFSET], bSD, bWE, SHAPE, [DATA1], [DATA2]                    ; 1st line - DB/USER
        # ; iSEC, TYPE, SNAME, [OFFSET], bSD, bWE, SHAPE, BLT, D1, ..., D8, iCEL              ; 1st line - VALUE
        # ;       AREA, ASy, ASz, Ixx, Iyy, Izz                                               ; 2nd line
        # ;       CyP, CyM, CzP, CzM, QyB, QzB, PERI_OUT, PERI_IN, Cy, Cz                     ; 3rd line
        # ;       Y1, Y2, Y3, Y4, Z1, Z2, Z3, Z4, Zyy, Zzz                                    ; 4th line
        # ; iSEC, TYPE, SNAME, [OFFSET], bSD, bWE, SHAPE, ELAST, DEN, POIS, POIC, SF, THERMAL ; 1st line - SRC
        # ;       D1, D2, [SRC]                                                               ; 2nd line
        # ; iSEC, TYPE, SNAME, [OFFSET], bSD, bWE, SHAPE, 1, DB, NAME1, NAME2, D1, D2         ; 1st line - COMBINED
        # ; iSEC, TYPE, SNAME, [OFFSET], bSD, bWE, SHAPE, 2, D11, D12, D13, D14, D15, D21, D22, D23, D24
        # ; iSEC, TYPE, SNAME, [OFFSET2], bSD, bWE, SHAPE, iyVAR, izVAR, STYPE                ; 1st line - TAPERED
        # ;       DB, NAME1, NAME2                                                            ; 2nd line(STYPE=DB)
        # ;       [DIM1], [DIM2]                                                              ; 2nd line(STYPE=USER)
        # ;       D11, D12, D13, D14, D15, D16, D17, D18                                      ; 2nd line(STYPE=VALUE)
        # ;       AREA1, ASy1, ASz1, Ixx1, Iyy1, Izz1                                         ; 3rd line(STYPE=VALUE)
        # ;       CyP1, CyM1, CzP1, CzM1, QyB1, QzB1, PERI_OUT1, PERI_IN1, Cy1, Cz1           ; 4th line(STYPE=VALUE)
        # ;       Y11, Y12, Y13, Y14, Z11, Z12, Z13, Z14, Zyy1, Zyy2                          ; 5th line(STYPE=VALUE)
        # ;       D21, D22, D23, D24, D25, D26, D27, D28                                      ; 6th line(STYPE=VALUE)
        # ;       AREA2, ASy2, ASz2, Ixx2, Iyy2, Izz2                                         ; 7th line(STYPE=VALUE)
        # ;       CyP2, CyM2, CzP2, CzM2, QyB2, QzB2, PERI_OUT2, PERI_IN2, Cy2, Cz2           ; 8th line(STYPE=VALUE)
        # ;       Y21, Y22, Y23, Y24, Z21, Z22, Z23, Z24, Zyy2, Zzz2                          ; 9th line(STYPE=VALUE)
        # ; iSEC, TYPE, SNAME, [OFFSET], bSD, bWE, SHAPE                                      ; 1st line - COMPOSITE-B
        # ;       Hw, tw, B1, Bf1, tf1, B2, Bf2, tf2                                          ; 2nd line
        # ;       [SHAPE-NUM], [STIFF-SHAPE], [STIFF-POS] (1~4)                               ; 3rd line
        # ;       SW, GN, CTC, Bc, Tc, Hh, EsEc, DsDc, Ps, Pc, TsTc, bMulti, Elong, Esh       ; 4th line
        # ; iSEC, TYPE, SNAME, [OFFSET], bSD, bWE, SHAPE                                      ; 1st line - COMPOSITE-I
        # ;       Hw, tw, B1, tf1, B2, tf2                                                    ; 2nd line
        # ;       [SHAPE-NUM], [STIFF-SHAPE], [STIFF-POS] (1~2)                               ; 3rd line
        # ;       SW, GN, CTC, Bc, Tc, Hh, EsEc, DsDc, Ps, Pc, TsTc, bMulti, Elong, Esh       ; 4th line
        # ; iSEC, TYPE, SNAME, [OFFSET], bSD, bWE, SHAPE                                      ; 1st line - COMPOSITE-TUB
        # ;       Hw, tw, B1, Bf1, tf1, B2, Bf2, tf2, Bf3, tfp                                ; 2nd line
        # ;       [SHAPE-NUM], [STIFF-SHAPE], [STIFF-POS] (1~3)                               ; 3rd line
        # ;       SW, GN, CTC, Bc, Tc, Hh, EsEc, DsDc, Ps, Pc, TsTc, bMulti, Elong, Esh       ; 4th line
        # ; [DATA1] : 1, DB, NAME or 2, D1, D2, D3, D4, D5, D6, D7, D8, D9, D10
        # ; [DATA2] : CCSHAPE or iCEL or iN1, iN2
        # ; [SRC]  : 1, DB, NAME1, NAME2 or 2, D1, D2, D3, D4, D5, D6, D7, D8, D9, D10, iN1, iN2
        # ; [DIM1], [DIM2] : D1, D2, D3, D4, D5, D6, D7, D8
        # ; [OFFSET] : OFFSET, iCENT, iREF, iHORZ, HUSER, iVERT, VUSER
        # ; [OFFSET2]: OFFSET, iCENT, iREF, iHORZ, HUSERI, HUSERJ, iVERT, VUSERI, VUSERJ
        # ; [SHAPE-NUM]: SHAPE-NUM, POS, STIFF-NUM1, STIFF-NUM2, STIFF-NUM3, STIFF-NUM4
        # ; [STIFF-SHAPE]: SHAPE-NUM, for(SHAPE-NUM) { NAME, SIZE1~8 }
        # ; [STIFF-POS]: STIFF-NUM, for(STIFF-NUM) { SPACING, iSHAPE, bCALC }
        # -----------------------------------------------------------------------------------------------------------

        sections = []
        for k, v in self.__sectionShapes.items():
            strFormat = "{idShape:>8d},DBUSER,{name:>16s}, CC, 0, 0, 0, 0, 0, 0, YES, NO, SB, 2,{width:>8.3f},{height:>8.3f},0,0,0,0,0,0,0,0"
            formatted = strFormat.format(
                idShape=k,
                name=v["name"],
                width=v["dimensions"][0],
                height=v["dimensions"][1],
            )
            sections.append(formatted)

            # --------------------------------
        # *STLDCASE    ; Static Load Cases
        # ; LCNAME, LCTYPE, DESC
        # --------------------------------
        loadCases = []
        for k in self.__loadCase.keys():
            v = self.__loadCase[k]
            strFormat = "{LCNAME:>8s},{LCTYPE:>3s},{DESC:>40s}"
            formatted = strFormat.format(LCNAME=k, LCTYPE=v[0], DESC=v[1])
            loadCases.append(formatted)

        # --------------------------------------------
        # *CONSTRAINT    ; Supports
        # ; NODE_LIST, CONST(Dx,Dy,Dz,Rx,Ry,Rz), GROUP
        # --------------------------------------------
        nodeContraints = []
        if len(self.__nodeSupports) > 0:
            nodeContraints.append("*CONSTRAINT")
            for k, v in self.__nodeSupports.items():
                strFormat = "{tag:d},{TX:d}{TY:d}{TZ:d}{RX:d}{RY:d}{RZ:d}"
                formatted = strFormat.format(
                    tag=k, TX=v[0], TY=v[1], TZ=v[2], RX=v[3], RY=v[4], RZ=v[5]
                )
                nodeContraints.append(formatted)

        # -----------------------------------------------------------------------------------------------------------------
        # *SPRING    ; Point Spring Supports
        # ; NODE_LIST, Type, F_SDx, F_SDy, F_SDz, F_SRx, F_SRy, F_SRz, SDx, SDy, SDz, SRx, SRy, SRz ...
        # ;                  DAMPING, Cx, Cy, Cz, CRx, CRy, CRz, GROUP, [DATA1]                                ; LINEAR
        # ; NODE_LIST, Type, Direction, Vx, Vy, Vz, Stiffness, GROUP, [DATA1]                                  ; COMP, TENS
        # ; NODE_LIST, Type, Direction, Vx, Vy, Vz, FUNCTION, GROUP, [DATA1]                                   ; MULTI
        # ; [DATA1] EFFAREA, Kx, Ky, Kz
        # -----------------------------------------------------------------------------------------------------------------
        springs = []
        if len(self.__nodeSprings) > 0:
            springs.append("*SPRING")
            for k, v in self.__nodeSprings.items():
                for kk, vv in v.items():
                    if kk == "COMP":
                        Type = "COMP"
                    elif kk == "TENS":
                        Type = "TENS"
                    elif kk == "LINEAR":
                        Type = "LINEAR"
                    else:
                        raise Ex.EXAExceptions(
                            "(EXAStructuralModel)-0026", "key error in Type", kk
                        )
                    for kkk, vvv in vv.items():
                        if kkk == "DX+":
                            Direction = 0
                        elif kkk == "DX-":
                            Direction = 1
                        elif kkk == "DY+":
                            Direction = 2
                        elif kkk == "DY-":
                            Direction = 3
                        elif kkk == "DZ+":
                            Direction = 4
                        elif kkk == "DZ-":
                            Direction = 5
                        else:
                            raise Ex.EXAExceptions(
                                "(EXAStructuralModel)-0026",
                                "key error in Direction",
                                kkk,
                            )
                        strFormat = "{tag:d},{Type:s},{Direction:d},0,0,0,{Stiffness:.6f}, , 0, 0, 0, 0, 0"
                        formatted = strFormat.format(
                            tag=k, Type=Type, Direction=Direction, Stiffness=vvv
                        )
                        springs.append(formatted)
            springs.append("")

        # -----------------
        # *USE-STLD, <name>
        # -----------------
        loadCasesUsed = []
        for k, v in self.__loads.items():
            strFormat = "*USE-STLD, {key:s}"
            formatted = strFormat.format(key=k)
            loadCasesUsed.append(formatted)
            loadCasesUsed.append("")
            for kk, vv in v.items():
                if kk == "frameLoad":
                    loadCasesUsed.append("*BEAMLOAD")
                    # *BEAMLOAD    ; Element Beam Loads
                    # ; ELEM_LIST, CMD, TYPE, DIR, bPROJ, [ECCEN], [VALUE], GROUP
                    # ; ELEM_LIST, CMD, TYPE, TYPE, DIR, VX, VY, VZ, bPROJ, [ECCEN], [VALUE], GROUP
                    # ; [VALUE]       : D1, P1, D2, P2, D3, P3, D4, P4
                    # ; [ECCEN]       : bECCEN, ECCDIR, I-END, J-END, bJ-END
                    # ; [ADDITIONAL]  : bADDITIONAL, ADDITIONAL_I-END, ADDITIONAL_J-END, bADDITIONAL_J-END
                    for kkk, vvv in vv.items():

                        for kkkk, vvvv in vvv.items():
                            if kkkk == "force":
                                tp = "UNILOAD"
                            elif kkkk == "moment":
                                tp = "UNIMOMENT"
                            else:
                                raise Ex.EXAExceptions(
                                    "(EXAStructuralModel)-0026", "key error in tp", kkkk
                                )

                            if "GCX1" in vvvv.keys() and "GCX2" in vvvv.keys():
                                strFormat = "{tag: d}, BEAM   , {tp}, GX, NO , NO, aDir[1], , , , {GCX1D:.6f}, {GCX1P:.6f}, {GCX2D:.6f}, {GCX2P:.6f}, 0, 0, 0, 0, , NO, 0, 0, NO,"
                                formatted = strFormat.format(
                                    tag=kkk,
                                    tp=tp,
                                    GCX1D=vvvv["GCX1"][0],
                                    GCX1P=vvvv["GCX1"][1],
                                    GCX2D=vvvv["GCX2"][0],
                                    GCX2P=vvvv["GCX2"][1],
                                )
                                loadCasesUsed.append(formatted)
                            if "GCY1" in vvvv.keys() and "GCY2" in vvvv.keys():
                                strFormat = "{tag: d}, BEAM   , {tp}, GY, NO , NO, aDir[1], , , , {GCY1D:.6f}, {GCY1P:.6f}, {GCY2D:.6f}, {GCY2P:.6f}, 0, 0, 0, 0, , NO, 0, 0, NO,"
                                formatted = strFormat.format(
                                    tag=kkk,
                                    tp=tp,
                                    GCY1D=vvvv["GCY1"][0],
                                    GCY1P=vvvv["GCY1"][1],
                                    GCY2D=vvvv["GCY2"][0],
                                    GCY2P=vvvv["GCY2"][1],
                                )
                                loadCasesUsed.append(formatted)
                            if "GCZ1" in vvvv.keys() and "GCZ2" in vvvv.keys():
                                strFormat = "{tag: d}, BEAM   , {tp}, GZ, NO , NO, aDir[1], , , , {GCZ1D:.6f}, {GCZ1P:.6f}, {GCZ2D:.6f}, {GCZ2P:.6f}, 0, 0, 0, 0, , NO, 0, 0, NO,"
                                formatted = strFormat.format(
                                    tag=kkk,
                                    tp=tp,
                                    GCZ1D=vvvv["GCZ1"][0],
                                    GCZ1P=vvvv["GCZ1"][1],
                                    GCZ2D=vvvv["GCZ2"][0],
                                    GCZ2P=vvvv["GCZ2"][1],
                                )
                                loadCasesUsed.append(formatted)

                    loadCasesUsed.append("")

                elif kk == "nodeLoad":
                    loadCasesUsed.append("*CONLOAD")
                    # *CONLOAD    ; Nodal Loads
                    # ; NODE_LIST, FX, FY, FZ, MX, MY, MZ, GROUP
                    for kkk, vvv in vv.items():
                        strFormat = "{tag: d}, {FX:.6f}, {FY:.6f}, {FZ:.6f}, {MX:.6f}, {MY:.6f}, {MZ:.6f},"
                        formatted = strFormat.format(
                            tag=kkk,
                            FX=vvv["NCS"][0],
                            FY=vvv["NCS"][1],
                            FZ=vvv["NCS"][2],
                            MX=vvv["NCS"][3],
                            MY=vvv["NCS"][4],
                            MZ=vvv["NCS"][5],
                        )
                        loadCasesUsed.append(formatted)

                    loadCasesUsed.append("")

                elif kk == "selfWeight":
                    loadCasesUsed.append("*SELFWEIGHT")
                    strFormat = "{DIRX: .6f}, {DIRY: .6f}, {DIRZ: .6f},"
                    formatted = strFormat.format(DIRX=vv[0], DIRY=vv[1], DIRZ=vv[2])
                    loadCasesUsed.append(formatted)
                    loadCasesUsed.append("")
                else:
                    raise Ex.EXAExceptions(
                        "(EXAStructuralModel)-0025", "key error in loadCasesUsed", kk
                    )

            loadCasesUsed.append(
                f"; End of data for load case {k} -------------------------"
            )

        output = template.render(
            udm="KN,M, BTU, C",
            nodes=nodes,
            elements=frames,
            loadCases=loadCases,
            loadCasesUsed=loadCasesUsed,
            contraints=nodeContraints,
            springs=springs,
            sections=sections,
        )

        # print(output)

        with open(fileName + ".mgt", "w") as f:
            f.write(output)

    def save(self, fileName: str, fileType: str=".msh"):
        print("Write ", fileName + fileType)
        gmsh.option.setNumber("Mesh.SaveAll", 1)
        if fileType == ".msh":
            gmsh.write(fileName + fileType)
            # logGmshFile(fileName + fileType)
        elif fileType == ".med":
            gmsh.write(fileName + fileType)
        elif fileType == ".mgt":
            self.__save_mgt(fileName)
        else:
            self.__log("ERR", f"File type unknoun *{fileType}*. None file saved !!!")

    def getCad(self):
        return self.__cad

    @staticmethod
    def getModeler():
        return gmsh_model

    def addModel(self, name: str) -> bool:
        if len(name) == 0:
            self.__log("ERR", "Arg must be a str not null !!!")
            return False
        else:
            if name in self.__models:
                self.__log("ERR", f"Model *{name}* is present !!!")
                return False
            else:
                gmsh_model.add(name)
                self.__models.append(name)
                self.__log("INF", f"Model *{name}* added !!!")
                return True

    def getModels(self) -> List[str]:
        return self.__models

    def setCurrentModel(self, name: str) -> bool:
        if name not in self.__models:
            self.__log("ERR", f"Model *{name}* not present !!!")
            return False
        else:
            gmsh_model.setCurrent(name)
            return True

    def addNodesToGroup(self, tags: List[int], tagGroup: int, physicalName: str=""):
        self.__addElementToGroup(tags, tagGroup, Dim.DIM_0D, physicalName)

    @staticmethod
    def __addElementToGroup(
        tags: List[int], tagGroup: int, dim: Dim, physicalName: str = ""
    ):
        gmsh_model.addPhysicalGroup(dim.value, tags, tagGroup)
        if len(physicalName) > 0:
            gmsh_model.setPhysicalName(dim.value, tagGroup, physicalName)

    @staticmethod
    def renumberFramesByNodeCoords(macroTags: List[Union[int,uint64]], dirRule: Literal['+X', '+Y', '+Z', '-X', '-Y', '-Z']):
        oldFrameTags = FEModel.getFramesFromMacroTags(macroTags)
        oldNodesTags = FEModel.getFrameNodesFromMacro(macroTags)

        idxCoord = ['X', 'Y', 'Z'].index(dirRule[1])
        oldMediumCoord = [0] * len(oldFrameTags)
        assert isinstance(oldNodesTags, list)

        for i, t in enumerate(oldNodesTags):
            oldMediumCoord[i] = (FEModel.nodeCoords(t[0])[idxCoord] + FEModel.nodeCoords(t[0])[idxCoord]) / 2

        oldTagCoords = zip(oldFrameTags, oldMediumCoord)

        rev = False
        if dirRule[0] == '-':
            rev = True

        oldTagCoordsSorted = sorted(oldTagCoords, key=lambda tc: tc[1], reverse=rev)

        newFrameTags = [0] * len(oldTagCoordsSorted)
        for i, t in enumerate(oldTagCoordsSorted):
            newFrameTags[i] = oldTagCoordsSorted[i][0]

        return oldFrameTags, newFrameTags

    @staticmethod
    def renumberArrange(oldMacroTags: List[List[int]], newMacroTags: List[List[int]]):
        compactOldTags = oldMacroTags[0]
        for i in range(len(oldMacroTags)-1):
            compactOldTags.extend(oldMacroTags[i+1])

        compactNewTags = newMacroTags[0]
        for i in range(len(newMacroTags)-1):
            compactNewTags.extend(newMacroTags[i+1])

        gmsh.model.mesh.renumberElements(compactOldTags, compactNewTags)

    def addFramesToGroup(self, tags: List[int], tagGroup: int, physicalName: str = "") -> None:
        """
        At macro element level adds frame to group.

        Args:
            tags ():
            tagGroup ():
            physicalName ():
        """
        self.__addElementToGroup(tags, tagGroup, Dim.DIM_1D, physicalName)
        for t in self.getFramesFromMacroTags(tags):
            if self.__framesPhysicalNames.get(t) is None:
                self.__framesPhysicalNames[t] = [physicalName]
            else:
                self.__framesPhysicalNames[t].append(physicalName)


    def getPhysicalNamesForFrame(self, tag):
        return self.__framesPhysicalNames.get(tag)

    def addNode(self,
                x: Union[int, float], y: Union[int, float], z: Union[int, float],
                tag: int=-1, group: bool=False, tagGroup: int=-1) -> int:
        tag = self.__cad.addPoint(x, y, z, meshSize=0.0, tag=tag)
        self.__cad.synchronize()
        if group:
            gmsh_model.addPhysicalGroup(0, [tag], tag=tagGroup)
            # self.__cad.synchronize()
        return tag

    def addFrame(self, tagStart: int, tagEnd: int, tag: int=-1, group: bool=False, tagGroup: int=-1, nb: int=1) -> int:
        tag = self.__cad.addLine(tagStart, tagEnd, tag=tag)
        self.__cad.synchronize()

        # mesh frame to have underlying mesh
        mesh = gmsh_model.mesh
        mesh.setTransfiniteCurve(tag, nb + 1)
        mesh.generate(1)

        if group:
            gmsh_model.addPhysicalGroup(1, [tag], tag=tagGroup)
            # self.__cad.synchronize()
        return tag

    def addPlateQuad(
        self,
        tagLine_1: int,
        tagLine_2: int,
        tagLine_3: int,
        tagLine_4: int,
        tag: int = -1,
        group: bool = False,
        tagGroup: int = -1,
        elType: Elem = Elem.ELEM_QUAD,
    ):
        wireId = self.__cad.addCurveLoop([tagLine_1, tagLine_2, tagLine_3, tagLine_4])
        plateId = self.__cad.addPlaneSurface(wireTags=[wireId], tag=tag)
        self.__cad.synchronize()
        mesh = gmsh_model.mesh
        mesh.setTransfiniteSurface(tag=plateId, arrangement="left")
        if elType == Elem.ELEM_QUAD:
            mesh.setRecombine(dim=2, tag=plateId)

        mesh.generate(2)

        if group:
            gmsh_model.addPhysicalGroup(2, [plateId], tag=tagGroup)
        return tag

    def addPlateQuadsToGroup(self, tags, tagGroup, physicalName=""):
        self.__addElementToGroup(tags, tagGroup, Dim.DIM_2D, physicalName)

    def addArc(self, tagStart, tagMid, tagEnd, tag=-1, group=False, tagGroup=-1, nb=10):
        tag = self.__cad.addCircleArc(tagStart, tagMid, tagEnd, tag=tag)
        self.__cad.synchronize()

        # mesh frame to have underlying mesh
        mesh = gmsh_model.mesh
        mesh.setTransfiniteCurve(tag, nb + 1)
        mesh.generate(1)

        if group:
            gmsh_model.addPhysicalGroup(1, [tag], tag=tagGroup)
            # self.__cad.synchronize()
        return tag


    def assignGroupTreeToPhysical(self, treePath: List[str | None], physicalName: str):
        if len(treePath) == 0:
            raise ValueError(f'treePath has len 0 !!!')

        if treePath[0] is None:
            raise ValueError(f'treePath starts with None type !!!')

        if physicalName == "":
            raise ValueError(f'physicalName has len 0 !!!')

        pn = self.getPhisicalNames()
        if not physicalName in pn:
            raise ValueError(f'physicalName given "{physicalName}" not in {pn} !!!')


        # check if None type are compacted at end of list
        hashStr = ""
        isNone = False
        for cols in treePath:
            if isNone:
                if cols is not None:
                    raise ValueError(f'None type within not none type !!!')
            if cols is None:
                isNone = True
            else:
                hashStr += cols + "/"

        self.__groupTree[hashStr] = physicalName

    def getGroupTreeToPhysical(self):
        return self.__groupTree


    def __str__(self):
        s = f"Model Description: {self.__description:s}"
        return s



import pprint
from typing import Any, Dict, List, Union, Literal

import pymongo

# importing ObjectId from bson library
from bson.objectid import ObjectId
from dotenv.variables import Literal
from pydantic_settings import BaseSettings
from pymongo.database import Collection, Database

from pycivil.EXAUtils.logging import log


class DatabaseSettings(BaseSettings):
    """Database settings fetched via environment variables."""

    host: str = "localhost"
    port: int = 27017
    username: Union[str, None] = None
    password: Union[str, None] = None

    class Config:
        """set the prefix for the env vars."""

        env_prefix = "DB_"


class DbManager:
    def __init__(self):
        self.__config = DatabaseSettings()
        # self.__host: str = 'localhost:27017'
        # self.__host: str = '127.0.0.1:27017'
        self.__timeOut: int = 6000
        self.__client: Union[pymongo.MongoClient[Dict[str, Any]], None] = None
        self.__db: Union[Database[dict[str, Any]], None] = None
        self.__cl: Union[Collection[dict[str, Any]], None] = None
        self.__ll: Literal[0, 1, 2, 3] = 3

    def __del__(self):
        if self.__client is None:
            return
        try:
            self.__client.close()
            log("INF", "Closed connection to server", self.__ll)
        except Exception:
            log("ERR", "I cannot close connection to server !!!", self.__ll)

    def connect(self) -> bool:
        if self.__client is not None:
            return True
        host = self.__config.host
        port = self.__config.port
        try:
            self.__client = pymongo.MongoClient(
                host=host,
                port=port,
                username=self.__config.username,
                password=self.__config.password,
                serverSelectionTimeoutMS=self.__timeOut,
                uuidRepresentation="standard",
            )
            log("INF", "Connection to server created, checking...", self.__ll)
            self.__client.server_info()
        except Exception as ex:
            log("ERR", f"Can't connect to database!!! {ex}", self.__ll)
            log(
                "ERR",
                f"host: {host}:{port} - timeout: {self.__timeOut}",
                self.__ll,
            )
            return False
        return True

    def newDb(self, dbName: str) -> bool:
        if self.__client is None:
            log("ERR", "Client not available !!!", self.__ll)
            return False
        if dbName in self.__client.list_database_names():
            log("ERR", f"dbName <{dbName}> yet unavailable !!!", self.__ll)
            return False
        self.__db = self.__client[dbName]
        self.__cl = self.__db["fuck_collection"]
        self.__cl.insert_one({})
        log(
            "INF",
            "after insert newDb databases: {}".format(
                self.__client.list_database_names()
            ),
            self.__ll,
        )
        return True

    def existsDb(self, dbName: str) -> bool:
        if self.__client is None:
            log("ERR", "Client not available !!!", self.__ll)
            return False
        return dbName in self.__client.list_database_names()

    def setCurrentDb(self, dbName: str) -> bool:
        if self.__client is not None:
            if dbName in self.__client.list_database_names():
                self.__db = self.__client[dbName]
                log("INF", f"dbName <{dbName}> make current", self.__ll)
                return True
            else:
                log("ERR", f"dbName <{dbName}> not available !!!", self.__ll)
                return False
        else:
            log("ERR", "Client not available !!!", self.__ll)
            return False

    def dropDb(self, dbName: str) -> bool:
        if self.__client is not None:
            if dbName in self.__client.list_database_names():
                self.__client.drop_database(dbName)
                log("INF", f"dbName <{dbName}> removed", self.__ll)
                log(
                    "INF",
                    f"databases: {self.__client.list_database_names()}",
                    self.__ll,
                )
                return True
            else:
                log("ERR", f"dbName <{dbName}> not available !!!", self.__ll)
                return False
        else:
            log("ERR", "Client not available !!!", self.__ll)
            return False

    def newCollection(self, collectionName: str) -> bool:
        if self.__db is not None:
            if collectionName not in self.__db.list_collection_names():
                self.__cl = self.__db[collectionName]
                self.__cl.insert_one({})
                log(
                    "INF",
                    f"collections: {self.__db.list_collection_names()}",
                    self.__ll,
                )
                return True
            else:
                log(
                    "ERR",
                    f"collection <{collectionName}> yet unavailable",
                    self.__ll,
                )
                return False
        else:
            log("ERR", "Db not available !!!", self.__ll)
            return False

    def setCurrentCollection(self, collectionName: str) -> bool:
        if self.__db is not None:
            if collectionName in self.__db.list_collection_names():
                self.__cl = self.__db[collectionName]
                log(
                    "INF",
                    f"collection <{collectionName}> make current",
                    self.__ll,
                )
                return True
            else:
                log(
                    "ERR",
                    f"collection <{collectionName}> not available !!!",
                    self.__ll,
                )
                return False
        else:
            log("ERR", "Client not available !!!", self.__ll)
            return False

    def newDocument(self, doc: Dict[str, Any]) -> str:
        if self.__cl is not None:
            id = self.__cl.insert_one(doc).inserted_id
            log("INF", f"doc added with id <{str(id)}>", self.__ll)
            return str(id)
        else:
            log("ERR", "None current collection available !!!", self.__ll)
            return ""

    def findOne(self, query: Dict[str, Any]) -> str:
        if self.__cl is not None:
            doc = self.__cl.find_one(query)
            if isinstance(doc, dict):
                log("INF", "doc found with id <{}>".format(str(doc["_id"])), self.__ll)
                return str(doc["_id"])
            else:
                log("ERR", "doc not found !!!", self.__ll)
                return ""
        else:
            log("ERR", "None current collection available !!!", self.__ll)
            return ""

    def getAllDocsFromCollection(self) -> Union[List[Dict], None]:
        if self.__cl is not None:
            return list(self.__cl.find())
        else:
            log("ERR", "None current collection available !!!", self.__ll)
            return None

    def replaceOne(self, query: Dict[str, Any], doc: Dict[str, Any]) -> bool:
        if self.__cl is None:
            log("ERR", "No current collection available !!!", self.__ll)
            return False
        mc = self.__cl.replace_one(query, doc).matched_count
        if mc == 0:
            log("INF", f"no document updated with query <{query}>", self.__ll)
            return False
        log("INF", "Doc updated", self.__ll)
        return True

    def findById(self, id: str) -> Dict[str, Any]:
        try:
            objectId = ObjectId(id)
        except Exception:
            log(
                "ERR",
                f"id <{id}> not compliant with possibilities given by QObject class!!!",
                self.__ll,
            )
            return {}

        if self.__cl is None:
            log("ERR", "No current collection available !!!", self.__ll)
            return {}
        doc = self.__cl.find_one({"_id": objectId})
        if isinstance(doc, dict):
            log("INF", "doc found by id <{}>".format(str(doc["_id"])), self.__ll)
            return doc
        log("ERR", "doc not found !!!", self.__ll)
        return {}

    def deleteById(self, id: str) -> int:
        try:
            objectId = ObjectId(id)
        except Exception:
            log(
                "ERR",
                f"id <{id}> not compliant with possibilities given by QObject class!!!",
                self.__ll,
            )
            return 0

        if self.__cl is None:
            log("ERR", "No current collection available !!!", self.__ll)
            return 0
        count = self.__cl.delete_one({"_id": objectId}).deleted_count
        if count > 0:
            log("INF", f"doc removed by id <{id}>", self.__ll)
            return count
        log("ERR", f"doc did not removed by id <{id}>", self.__ll)
        return 0

    def __str__(self) -> str:
        lines = []
        if self.__client is not None:
            lines.append(f"Database names: {self.__client.list_database_names()}")
        if self.__db is not None:
            lines.append(f"Collection names: {self.__db.list_collection_names()}")
        if self.__cl is not None:
            lines.append(f"Documents nb: {self.__cl.count()}")
        return "\n".join(lines)

    def printCollection(self) -> None:
        if self.__cl is None:
            log("ERR", "None current collection available !!!", self.__ll)
            return
        cursor = self.__cl.find({})
        for document in cursor:
            pprint.pprint(document, indent=4)



from enum import Enum
from typing import Any, Union

from pydantic import BaseModel, field_validator, model_validator, ConfigDict
from typing_extensions import Self

from pycivil.EXAStructural.lawcodes.codeNTC2018 import (
    rcSectionBeamAreaMin,
    rcSectionBeamShearAreaMin,
    rcSectionBeamShearStepsMax,
    rebarArea,
    rebarNumber,
)
from pycivil.EXAStructural.codes import Code
from pycivil.EXAStructural.materials import Concrete, ConcreteSteel
from pycivil.EXAUtils.solver import Solver


class BeamMinimumAreaInput(BaseModel):
    elementDescr: str = "Nome elemento"
    keyCode: str = "NTC2018"
    steelClass: str = "B450C"
    concreteClass: str = "C20/25"
    wEl: float = 300.0
    hEl: float = 600.0
    bt: float = 300.0
    rebarD: float = 12
    cover: float = 40.0
    bMin: float = 300.0
    stirrupD: float = 8
    rebarDComp: float = 12
    nbLegDirX: float = 2

    model_config = ConfigDict(extra="forbid")

    @field_validator("elementDescr", "keyCode", "steelClass", "concreteClass")
    @classmethod
    def must_be_string(cls, v: Any) -> Any:
        assert isinstance(v, str)
        return v

    @field_validator(
        "wEl",
        "hEl",
        "bt",
        "rebarD",
        "cover",
        "bMin",
        "stirrupD",
        "rebarDComp",
        "nbLegDirX",
        mode="before"
    )
    @classmethod
    def greater_than_zero(cls, v: Any) -> Any:
        assert isinstance(v, float)
        assert v > 0
        return v

    @model_validator(mode="after")
    def rebar_in_concrete(self) -> Self:
        if self.hEl - self.rebarD - self.cover < 0:
            raise ValueError("Concrete height")
        return self


class BeamMinimumAreaOutput(BaseModel):
    rebarNumber: int = 0
    stirrupStepMin: float = 0
    stirrupCompStepMin: float = 0


class BeamMinimumAreaLogs(BaseModel):
    clsClass: str = ""
    cls_fck: float = 0
    cls_fctm: float = 0
    steelClass: str = ""
    steel_fyk: float = 0
    heightUtil: float = 0
    areaUtil: float = 0
    minimumRebarAreaCrit1: float = 0
    minimumRebarAreaCrit2: float = 0
    minimumRebarArea: float = 0
    rebarAreaDisposed: float = 0
    minimumRebarAreaForElementLenght: float = 0
    minimumLegsForElementLenght: float = 0
    maxStepCrit1: float = 0
    maxStepCrit2: float = 0
    maxStepCrit3: float = 0
    maxStepCrit4: float = 0


class BeamMinimumArea(BaseModel):
    inputData: BeamMinimumAreaInput = BeamMinimumAreaInput()
    logsData: BeamMinimumAreaLogs = BeamMinimumAreaLogs()
    outputData: BeamMinimumAreaOutput = BeamMinimumAreaOutput()


class ConcreteMaterialInput(BaseModel):
    elementDescr: str = "Nome elemento"
    keyCode: str = "NTC2018"
    concreteClass: str = "C20/25"


class RebarMaterialInput(BaseModel):
    elementDescr: str = "Nome elemento"
    keyCode: str = "NTC2018"
    steelClass: str = "B450C"


class ConcreteMaterialOutput(BaseModel):
    value_Rck: float = 0
    value_fck: float = 0
    value_fcm: float = 0
    value_fctm: float = 0
    value_Ecm: float = 0
    value_gammac: float = 0
    value_fcd: float = 0
    value_alphacc: float = 0
    value_sigmaCar: float = 0
    value_sigmaQp: float = 0


class RebarMaterialOutput(BaseModel):
    value_fyk: float = 0
    value_ftk: float = 0
    value_Es: float = 0
    value_gammas: float = 0
    value_fyd: float = 0
    value_sigmaCar: float = 0


class ConcreteMaterial(BaseModel):
    inputData: ConcreteMaterialInput = ConcreteMaterialInput()
    outputData: ConcreteMaterialOutput = ConcreteMaterialOutput()


class RebarMaterial(BaseModel):
    inputData: RebarMaterialInput = RebarMaterialInput()
    outputData: RebarMaterialOutput = RebarMaterialOutput()


class SolverConcrete(Solver):
    def __init__(self, inputModel: ConcreteMaterialInput):
        if not isinstance(inputModel, ConcreteMaterialInput):
            raise ValueError
        self.__in = inputModel
        super().__init__(modelInput=inputModel)
        self.__out = ConcreteMaterialOutput()
        super()._setModelOutput(model=self.__out)

    def run(self, opt: Union[Enum, None] = None, **kwargs) -> bool:
        if opt is None:
            mat = Concrete()
            mat.setByCode(
                codeObj=Code(self.__in.keyCode), catstr=self.__in.concreteClass
            )
            self.__out.value_Rck = mat.get_Rck()
            self.__out.value_fck = mat.get_fck()
            self.__out.value_fcd = mat.cal_fcd()
            self.__out.value_fcm = mat.get_fcm()
            self.__out.value_fctm = mat.get_fctm()
            self.__out.value_Ecm = mat.get_Ecm()
            self.__out.value_gammac = mat.get_gammac()
            self.__out.value_alphacc = mat.get_alphacc()
            self.__out.value_sigmaCar = mat.get_sigmac_max_c()
            self.__out.value_sigmaQp = mat.get_sigmac_max_q()
            return True
        else:
            return False


class SolverSteelRebar(Solver):
    def __init__(self, inputModel: RebarMaterialInput):
        if not isinstance(inputModel, RebarMaterialInput):
            raise ValueError
        self.__in = inputModel
        super().__init__(modelInput=inputModel)
        self.__out = RebarMaterialOutput()
        super()._setModelOutput(model=self.__out)

    def run(self, opt: Union[Enum, None] = None, **kwargs) -> bool:
        mat = ConcreteSteel()
        mat.setByCode(codeObj=Code(self.__in.keyCode), catstr=self.__in.steelClass)
        self.__out.value_fyk = mat.get_fsy()
        self.__out.value_Es = mat.get_Es()
        self.__out.value_fyd = mat.cal_fyd()
        self.__out.value_ftk = mat.get_fuk()
        self.__out.value_gammas = mat.get_gammas()
        self.__out.value_sigmaCar = mat.get_sigmas_max_c()
        return True


class SolverBeamMinRebar(Solver):
    def __init__(self, inputModel: BeamMinimumAreaInput):
        if not isinstance(inputModel, BeamMinimumAreaInput):
            raise ValueError
        self.__in = inputModel
        super().__init__(modelInput=inputModel)
        self.__logs = BeamMinimumAreaLogs()
        super()._setModelLogs(model=self.__logs)
        self.__out = BeamMinimumAreaOutput()
        super()._setModelOutput(model=self.__out)

    def run(self, opt: Union[Enum, None] = None, **kwargs) -> bool:

        # Materials
        mat_steel = ConcreteSteel()
        mat_steel.setByCode(
            codeObj=Code(self.__in.keyCode), catstr=self.__in.steelClass
        )
        self.__logs.steelClass = mat_steel.catStr()
        self.__logs.steel_fyk = mat_steel.get_fsy()
        mat_concrete = Concrete()
        mat_concrete.setByCode(
            codeObj=Code(self.__in.keyCode), catstr=self.__in.concreteClass
        )
        self.__logs.clsClass = mat_concrete.catStr()
        self.__logs.cls_fck = mat_concrete.get_fck()
        self.__logs.cls_fctm = mat_concrete.get_fctm()

        # Geometry for main rebar
        self.__logs.heightUtil = self.__in.hEl - self.__in.cover - self.__in.rebarD / 2
        self.__logs.minimumRebarAreaCrit1 = rcSectionBeamAreaMin(
            self.__logs.cls_fck,
            self.__logs.steel_fyk,
            self.__in.bt,
            self.__logs.heightUtil,
            "c1",
        )
        self.__logs.minimumRebarAreaCrit2 = rcSectionBeamAreaMin(
            self.__logs.cls_fck,
            self.__logs.steel_fyk,
            self.__in.bt,
            self.__logs.heightUtil,
            "c2",
        )
        self.__logs.minimumRebarArea = rcSectionBeamAreaMin(
            self.__logs.cls_fck,
            self.__logs.steel_fyk,
            self.__in.bt,
            self.__logs.heightUtil,
        )
        self.__logs.areaUtil = self.__in.bt * self.__logs.heightUtil
        self.__out.rebarNumber = rebarNumber(
            self.__logs.minimumRebarArea, self.__in.rebarD
        )[0]
        self.__logs.rebarAreaDisposed = self.__out.rebarNumber * rebarArea(
            self.__in.rebarD
        )

        # Geometry for stirrups
        self.__logs.minimumRebarAreaForElementLenght = rcSectionBeamShearAreaMin(
            self.__in.bMin
        )
        self.__logs.minimumLegsForElementLenght = rebarNumber(
            self.__logs.minimumRebarAreaForElementLenght, self.__in.stirrupD
        )[0]
        self.__logs.maxStepCrit1 = (
            self.__in.nbLegDirX / self.__logs.minimumLegsForElementLenght * 1000
        )
        (
            self.__logs.maxStepCrit2,
            self.__logs.maxStepCrit3,
            self.__logs.maxStepCrit4,
        ) = rcSectionBeamShearStepsMax(self.__logs.heightUtil, self.__in.rebarDComp)
        self.__out.stirrupStepMin = min(
            [
                self.__logs.maxStepCrit3,
                self.__logs.maxStepCrit1,
                self.__logs.maxStepCrit2,
            ]
        )
        self.__out.stirrupCompStepMin = min(
            [self.__out.stirrupStepMin, self.__logs.maxStepCrit4]
        )
        return True
